/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 The Boeing Company
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

// 
// This script configures two nodes on an 802.11b physical layer, with
// 802.11b NICs in adhoc mode, and by default, sends one packet of 1000 
// (application) bytes to the other node.  The physical layer is configured
// to receive at a fixed RSS (regardless of the distance and transmit
// power); therefore, changing position of the nodes has no effect. 
//
// There are a number of command-line options available to control
// the default behavior.  The list of available command-line options
// can be listed with the following command:
// ./waf --run "wifi-simple-adhoc --help"
//
// For instance, for this configuration, the physical layer will
// stop successfully receiving packets when rss drops below -97 dBm.
// To see this effect, try running:
//
// ./waf --run "wifi-simple-adhoc --rss=-97 --numPackets=20"
// ./waf --run "wifi-simple-adhoc --rss=-98 --numPackets=20"
// ./waf --run "wifi-simple-adhoc --rss=-99 --numPackets=20"
//
// Note that all ns-3 attributes (not just the ones exposed in the below
// script) can be changed at command line; see the documentation.
//
// This script can also be helpful to put the Wifi layer into verbose
// logging mode; this command will turn on all wifi logging:
// 
// ./waf --run "wifi-simple-adhoc --verbose=1"
//
// When you are done, you will notice two pcap trace files in your directory.
// If you have tcpdump installed, you can try this:
//
// tcpdump -r wifi-simple-adhoc-0-0.pcap -nn -tt
//


#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/mobility-module.h"
#include "ns3/config-store-module.h"
#include "ns3/wifi-module.h"
#include "ns3/internet-module.h"

#include <iostream>
#include <fstream>
#include <vector>
#include <string>

using namespace ns3;


/* application */

class MyApp : public Application
{
public:

    MyApp();
    virtual ~MyApp();

    void Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate);
    void ChangeRate(DataRate newrate);

private:
    virtual void StartApplication(void);
    virtual void StopApplication(void);

    void ScheduleTx(void);
    void SendPacket(void);

    Ptr<Socket> m_socket;
    Address m_peer;
    uint32_t m_packetSize;
    uint32_t m_nPackets;
    DataRate m_dataRate;
    EventId m_sendEvent;
    bool m_running;
    uint32_t m_packetsSent;
};

MyApp::MyApp()
: m_socket(0),
m_peer(),
m_packetSize(0),
m_nPackets(0),
m_dataRate(0),
m_sendEvent(),
m_running(false),
m_packetsSent(0)
{
}

MyApp::~MyApp()
{
    m_socket = 0;
}

void
MyApp::Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate)
{
    m_socket = socket;
    m_peer = address;
    m_packetSize = packetSize;
    m_nPackets = nPackets;
    m_dataRate = dataRate;
}

void
MyApp::StartApplication(void)
{
    m_running = true;
    m_packetsSent = 0;
    m_socket->Bind();
    m_socket->Connect(m_peer);
    SendPacket();
}

void
MyApp::StopApplication(void)
{
    m_running = false;

    if (m_sendEvent.IsRunning())
    {
        Simulator::Cancel(m_sendEvent);
    }

    if (m_socket)
    {
        m_socket->Close();
    }
}

void
MyApp::SendPacket(void)
{
    Ptr<Packet> packet = Create<Packet> (m_packetSize);
    m_socket->Send(packet);

    if (++m_packetsSent < m_nPackets)
    {
        ScheduleTx();
    }
}

void
MyApp::ScheduleTx(void)
{
    if (m_running)
    {
        Time tNext(Seconds(m_packetSize * 8 / static_cast<double> (m_dataRate.GetBitRate())));
        m_sendEvent = Simulator::Schedule(tNext, &MyApp::SendPacket, this);
    }
}

void
MyApp::ChangeRate(DataRate newrate)
{
    m_dataRate = newrate;
    return;
}

/* application */



NS_LOG_COMPONENT_DEFINE("WifiSimpleAdhoc");

void ReceivePacket(Ptr<Socket> socket)
{
    while (socket->Recv())
    {
        NS_LOG_UNCOND("Received one packet!");
    }
}

static void GenerateTraffic(Ptr<Socket> socket, uint32_t pktSize,
        uint32_t pktCount, Time pktInterval)
{
    if (pktCount > 0)
    {
        socket->Send(Create<Packet> (pktSize));
        Simulator::Schedule(pktInterval, &GenerateTraffic,
                socket, pktSize, pktCount - 1, pktInterval);
    }
    else
    {
        socket->Close();
    }
}

#define PLABAN

int main(int argc, char *argv[])
{
#if 1
    LogComponentEnable("WifiPhy", LOG_LEVEL_ALL);
    LogComponentEnable("WifiPhy", LOG_PREFIX_FUNC);
    LogComponentEnable("WifiPhy", LOG_PREFIX_NODE);
    LogComponentEnable("WifiPhy", LOG_PREFIX_TIME);
#endif 

#if 1
    LogComponentEnable("AdhocWifiMac", LOG_LEVEL_ALL);
    LogComponentEnable("AdhocWifiMac", LOG_PREFIX_FUNC);
    LogComponentEnable("AdhocWifiMac", LOG_PREFIX_NODE);
    LogComponentEnable("AdhocWifiMac", LOG_PREFIX_TIME);
#endif 
    
#if 1
    LogComponentEnable("YansWifiPhy", LOG_LEVEL_ALL);
    LogComponentEnable("YansWifiPhy", LOG_PREFIX_FUNC);
    LogComponentEnable("YansWifiPhy", LOG_PREFIX_NODE);
    LogComponentEnable("YansWifiPhy", LOG_PREFIX_TIME);
#endif 

#if 1
    LogComponentEnable("YansWifiChannel", LOG_LEVEL_ALL);
    LogComponentEnable("YansWifiChannel", LOG_PREFIX_FUNC);
    LogComponentEnable("YansWifiChannel", LOG_PREFIX_NODE);
    LogComponentEnable("YansWifiChannel", LOG_PREFIX_TIME);
#endif 
   
#if 1
    LogComponentEnable("Socket", LOG_LEVEL_ALL);
    LogComponentEnable("Socket", LOG_PREFIX_FUNC);
    LogComponentEnable("Socket", LOG_PREFIX_NODE);
    LogComponentEnable("Socket", LOG_PREFIX_TIME);
#endif 
    
#if 1
    LogComponentEnable("WifiPhyStateHelper", LOG_LEVEL_ALL);
    LogComponentEnable("WifiPhyStateHelper", LOG_PREFIX_FUNC);
    LogComponentEnable("WifiPhyStateHelper", LOG_PREFIX_NODE);
    LogComponentEnable("WifiPhyStateHelper", LOG_PREFIX_TIME);
#endif 
    
#if 1
    LogComponentEnable("MacLow", LOG_LEVEL_ALL);
    LogComponentEnable("MacLow", LOG_PREFIX_FUNC);
    LogComponentEnable("MacLow", LOG_PREFIX_NODE);
    LogComponentEnable("MacLow", LOG_PREFIX_TIME);
#endif  
    

    

    std::string phyMode("DsssRate1Mbps");
    double rss = -80; // -dBm
    uint32_t packetSize = 1000; // bytes
    uint32_t numPackets = 1;
    double interval = 1.0; // seconds
    bool verbose = false;

    CommandLine cmd;

    cmd.AddValue("phyMode", "Wifi Phy mode", phyMode);
    cmd.AddValue("rss", "received signal strength", rss);
    cmd.AddValue("packetSize", "size of application packet sent", packetSize);
    cmd.AddValue("numPackets", "number of packets generated", numPackets);
    cmd.AddValue("interval", "interval (seconds) between packets", interval);
    cmd.AddValue("verbose", "turn on all WifiNetDevice log components", verbose);

    cmd.Parse(argc, argv);
    // Convert to time object
    Time interPacketInterval = Seconds(interval);

    // disable fragmentation for frames below 2200 bytes
    Config::SetDefault("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue("2200"));
    // turn off RTS/CTS for frames below 2200 bytes
    Config::SetDefault("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue("2200"));
    // Fix non-unicast data rate to be the same as that of unicast
    Config::SetDefault("ns3::WifiRemoteStationManager::NonUnicastMode", StringValue(phyMode));

    NodeContainer c;
    c.Create(2);

#ifdef PLABAN
    YansWifiPhyHelper wifiPhyHlpr = YansWifiPhyHelper::Default();
    YansWifiChannelHelper wifiChnlHlpr = YansWifiChannelHelper::Default();
    Ptr<YansWifiChannel> chanel = wifiChnlHlpr.Create();

    //Ptr<LogDistancePropagationLossModel> propLossModel = CreateObject<LogDistancePropagationLossModel>();
    //channel->SetPropagationLossModel(propLossModel);

    wifiPhyHlpr.SetChannel(chanel);

    WifiHelper wifiHlpr;
    if (verbose)
    {
        wifiHlpr.EnableLogComponents(); // Turn on all Wifi logging
    }
    wifiHlpr.SetStandard(WIFI_PHY_STANDARD_80211n_2_4GHZ); // i have changed it to 802.11n2.4GHz

    // Add a non-QoS upper mac, and disable rate control
    NqosWifiMacHelper wifiMacHlpr = NqosWifiMacHelper::Default();
    wifiHlpr.SetRemoteStationManager("ns3::ConstantRateWifiManager",
            "DataMode", StringValue(phyMode),
            "ControlMode", StringValue(phyMode));
    // Set it to adhoc mode
    wifiMacHlpr.SetType("ns3::AdhocWifiMac");

#else 

    // The below set of helpers will help us to put together the wifi NICs we want
    WifiHelper wifiHlpr;
    if (verbose)
    {
        wifiHlpr.EnableLogComponents(); // Turn on all Wifi logging
    }

    wifiHlpr.SetStandard(WIFI_PHY_STANDARD_80211b);

    YansWifiPhyHelper wifiPhyHlpr = YansWifiPhyHelper::Default();
    // This is one parameter that matters when using FixedRssLossModel
    // set it to zero; otherwise, gain will be added
    wifiPhyHlpr.Set("RxGain", DoubleValue(0));
    // ns-3 supports RadioTap and Prism tracing extensions for 802.11b
    wifiPhyHlpr.SetPcapDataLinkType(YansWifiPhyHelper::DLT_IEEE802_11_RADIO);
    YansWifiChannelHelper wifiChnlHlpr;
    wifiChnlHlpr.SetPropagationDelay("ns3::ConstantSpeedPropagationDelayModel");
    // The below FixedRssLossModel will cause the rss to be fixed regardless
    // of the distance between the two stations, and the transmit power
    wifiChnlHlpr.AddPropagationLoss("ns3::FixedRssLossModel", "Rss", DoubleValue(rss));
    wifiPhyHlpr.SetChannel(wifiChnlHlpr.Create());

    // Add a non-QoS upper mac, and disable rate control
    NqosWifiMacHelper wifiMacHlpr = NqosWifiMacHelper::Default();
    wifiHlpr.SetRemoteStationManager("ns3::ConstantRateWifiManager",
            "DataMode", StringValue(phyMode),
            "ControlMode", StringValue(phyMode));
    // Set it to adhoc mode
    wifiMacHlpr.SetType("ns3::AdhocWifiMac");
#endif 

    NetDeviceContainer devices = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, c);

    // Note that with FixedRssLossModel, the positions below are not 
    // used for received signal strength. 
    MobilityHelper mobility;
    Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();
    positionAlloc->Add(Vector(0.0, 0.0, 0.0));
    positionAlloc->Add(Vector(5.0, 0.0, 0.0));
    mobility.SetPositionAllocator(positionAlloc);
    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(c);

    InternetStackHelper internet;
    internet.Install(c);

    Ipv4AddressHelper ipv4;
    NS_LOG_INFO("Assign IP Addresses.");
    ipv4.SetBase("10.1.1.0", "255.255.255.0");
    Ipv4InterfaceContainer i = ipv4.Assign(devices);

    TypeId tid = TypeId::LookupByName("ns3::UdpSocketFactory");
    Ptr<Socket> recvSink = Socket::CreateSocket(c.Get(0), tid);
    InetSocketAddress local = InetSocketAddress(Ipv4Address::GetAny(), 80);
    recvSink->Bind(local);
    recvSink->SetRecvCallback(MakeCallback(&ReceivePacket));

    Ptr<Socket> source = Socket::CreateSocket(c.Get(1), tid);
    InetSocketAddress remote = InetSocketAddress(Ipv4Address("255.255.255.255"), 80);
    source->SetAllowBroadcast(true);
    source->Connect(remote);

    // Tracing
    wifiPhyHlpr.EnablePcap("wifi-simple-adhoc", devices);

    // Output what we are doing
    NS_LOG_UNCOND("Testing " << numPackets << " packets sent with receiver rss " << rss);

    Simulator::ScheduleWithContext(source->GetNode()->GetId(),
            Seconds(1.0), &GenerateTraffic,
            source, packetSize, numPackets, interPacketInterval);

    Simulator::Run();
    Simulator::Destroy();

    return 0;
}

