
#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/mobility-module.h"
#include "ns3/config-store-module.h"
#include "ns3/wifi-module.h"
#include "ns3/csma-module.h"
#include "ns3/internet-module.h"
#include <fstream>
#include <string>
#include "ns3/point-to-point-module.h"
#include "ns3/applications-module.h"
#include "ns3/flow-monitor-module.h"
#include "ns3/ipv4-global-routing-helper.h" 

#include <iostream>
#include <fstream>
#include <vector>
#include <string>

using namespace ns3;

static void RcvBySink(Ptr<const Packet> pkt, const Address &)
{
    std::cout << Simulator::Now().GetSeconds() << "\t\t\t\tpkt received, pkt Uid = " << pkt->GetUid() << "\n";
}

void ReceivePacket(Ptr<Socket> socket)
{
    while (socket->Recv())
    {
        NS_LOG_UNCOND("Received one packet!");
    }
}

static void GenerateTraffic(Ptr<Socket> socket, uint32_t pktSize,
        uint32_t pktCount, Time pktInterval)
{
    if (pktCount > 0)
    {
        socket->Send(Create<Packet> (pktSize));
        Simulator::Schedule(pktInterval, &GenerateTraffic,
                socket, pktSize, pktCount - 1, pktInterval);
    }
    else
    {
        socket->Close();
    }
}

class MyApp : public Application
{
public:

    MyApp();
    virtual ~MyApp();

    void Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate);
    void ChangeRate(DataRate newrate);

private:
    virtual void StartApplication(void);
    virtual void StopApplication(void);

    void ScheduleTx(void);
    void SendPacket(void);

    Ptr<Socket> m_socket;
    Address m_peer;
    uint32_t m_packetSize;
    uint32_t m_nPackets;
    DataRate m_dataRate;
    EventId m_sendEvent;
    bool m_running;
    uint32_t m_packetsSent;
};

MyApp::MyApp()
: m_socket(0),
m_peer(),
m_packetSize(0),
m_nPackets(0),
m_dataRate(0),
m_sendEvent(),
m_running(false),
m_packetsSent(0)
{
}

MyApp::~MyApp()
{
    m_socket = 0;
}

void
MyApp::Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate)
{
    m_socket = socket;
    m_peer = address;
    m_packetSize = packetSize;
    m_nPackets = nPackets;
    m_dataRate = dataRate;
}

void
MyApp::StartApplication(void)
{
    m_running = true;
    m_packetsSent = 0;
    m_socket->Bind();
    m_socket->Connect(m_peer);
    SendPacket();
}

void
MyApp::StopApplication(void)
{
    m_running = false;

    if (m_sendEvent.IsRunning())
    {
        Simulator::Cancel(m_sendEvent);
    }

    if (m_socket)
    {
        m_socket->Close();
    }
}

void
MyApp::SendPacket(void)
{
    Ptr<Packet> packet = Create<Packet> (m_packetSize);
    m_socket->Send(packet);

    if (++m_packetsSent < m_nPackets)
    {
        ScheduleTx();
    }
}

void
MyApp::ScheduleTx(void)
{
    if (m_running)
    {
        Time tNext(Seconds(m_packetSize * 8 / static_cast<double> (m_dataRate.GetBitRate())));
        m_sendEvent = Simulator::Schedule(tNext, &MyApp::SendPacket, this);
    }
}

void
MyApp::ChangeRate(DataRate newrate)
{
    m_dataRate = newrate;
    return;
}

static void
CwndChange(uint32_t oldCwnd, uint32_t newCwnd)
{
    std::cout << Simulator::Now().GetSeconds() << "\t" << newCwnd << "\n";
}

void
IncRate(Ptr<MyApp> app, DataRate rate)
{
    app->ChangeRate(rate);
    return;
}

class AP_STA_Cluster : public Object
{
public:
    int simulationTimeSec;
    int m_wifi_stationCount;
    int m_numberOfAPchannel;
    int m_numberOfNodeINcsmaAPcluster;
    
    CsmaHelper csmaHlpr;

    NodeContainer apCsmaNodeContainer;
    NodeContainer wifiApNodeContainer;
    NodeContainer wifiStaNodeContainer;

    NetDeviceContainer apDevicesContainer;
    NetDeviceContainer staDevicesContainer;

    Ipv4InterfaceContainer csmaIpv4IntfcCntnr;
    Ipv4InterfaceContainer apInterfaceContainer;
    Ipv4InterfaceContainer staInterfaceContainer;

    AP_STA_Cluster()
    {
        this->m_numberOfAPchannel = 3;
        this->m_numberOfNodeINcsmaAPcluster = 1 + this->m_numberOfAPchannel;
    }
            
    void setup(int _simulationTimeSec, int _m_wifi_stationCount, WifiHelper & wifiHlpr, YansWifiPhyHelper & wifiPhyHlpr,
            NqosWifiMacHelper & wifiMacHlpr, Ipv4AddressHelper & ipv4Hlpr, InternetStackHelper & g_internetStkHlpr)
    {
        this->simulationTimeSec = _simulationTimeSec;
        this->m_wifi_stationCount = _m_wifi_stationCount;
        this->apCsmaNodeContainer.Create(m_numberOfNodeINcsmaAPcluster);

        
        this->csmaHlpr.SetChannelAttribute("DataRate", StringValue("100Mbps")); // TODO:: change data rate
        this->csmaHlpr.SetChannelAttribute("Delay", TimeValue(NanoSeconds(6560))); // TODO: change delay time

        NetDeviceContainer csmaNetDevCntnr;
        csmaNetDevCntnr = csmaHlpr.Install(this->apCsmaNodeContainer); // all the APs will be connected in a CSMA network


        for (int i = 1; i < this->m_numberOfNodeINcsmaAPcluster; i++)
        {
            this->wifiApNodeContainer.Add(this->apCsmaNodeContainer.Get(i));
        }

        this->wifiStaNodeContainer.Create(this->m_wifi_stationCount);

        this->apDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, this->wifiApNodeContainer);
        this->staDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, this->wifiStaNodeContainer);

        /*Mobility  model start*/
        MobilityHelper mobility;

        mobility.SetPositionAllocator("ns3::GridPositionAllocator",
                "MinX", DoubleValue(0.0),
                "MinY", DoubleValue(0.0),
                "DeltaX", DoubleValue(5.0),
                "DeltaY", DoubleValue(10.0),
                "GridWidth", UintegerValue(3),
                "LayoutType", StringValue("RowFirst"));

        /*
        mobility.SetMobilityModel("ns3::RandomWalk2dMobilityModel",
                "Bounds", RectangleValue(Rectangle(-50, 50, -50, 50)));
        mobility.Install(wifiStaNodes);
         */

        mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
        mobility.Install(wifiApNodeContainer);
        mobility.Install(wifiStaNodeContainer);
        /*Mobility model end*/

        g_internetStkHlpr.Install(apCsmaNodeContainer);
        g_internetStkHlpr.Install(wifiStaNodeContainer);

        ipv4Hlpr.SetBase("10.1.1.0", "255.255.255.0");
        this->csmaIpv4IntfcCntnr = ipv4Hlpr.Assign(csmaNetDevCntnr);


        ipv4Hlpr.SetBase("10.1.2.0", "255.255.255.0");
        this->apInterfaceContainer = ipv4Hlpr.Assign(apDevicesContainer);
        this->staInterfaceContainer = ipv4Hlpr.Assign(staDevicesContainer);
    }

    void startTCPfromStaToAP(int _staID, uint32_t _pktSize, uint32_t _pktNumber)
    {
        uint32_t TCPpacketSize = _pktSize; // bytes
        uint32_t TCPnumPackets = _pktNumber;
        int staID = _staID;
        
        NS_ASSERT_MSG((0<=staID && staID<m_wifi_stationCount), "wifi stationID index out of bound !" );

        /* TCP Sink start*/
        uint16_t sinkPort = 8080;
        Address sinkAddress(InetSocketAddress(csmaIpv4IntfcCntnr.GetAddress(staID), sinkPort)); // interface of sink
        PacketSinkHelper packetSinkHelper("ns3::TcpSocketFactory", InetSocketAddress(Ipv4Address::GetAny(), sinkPort));
        ApplicationContainer sinkApps = packetSinkHelper.Install(apCsmaNodeContainer.Get(staID));
        sinkApps.Start(Seconds(0.));
        sinkApps.Stop(Seconds(100.));
        /* TCP Sink end*/

        /*TCP Source start*/
        Ptr<Socket> ns3TcpSocket = Socket::CreateSocket(wifiStaNodeContainer.Get(staID), TcpSocketFactory::GetTypeId());
        // Trace Congestion window
        ns3TcpSocket->TraceConnectWithoutContext("CongestionWindow", MakeCallback(&CwndChange));
        Config::ConnectWithoutContext("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&RcvBySink));

        // Create TCP application at n0
        Ptr<MyApp> app = CreateObject<MyApp> ();
        app->Setup(ns3TcpSocket, sinkAddress, TCPpacketSize, TCPnumPackets, DataRate("250Kbps"));
        wifiStaNodeContainer.Get(staID)->AddApplication(app);
        app->SetStartTime(Seconds(1.));
        app->SetStopTime(Seconds(100.));
    }
};

int main(int argc, char *argv[])
{
    bool verbose = false;
    //std::string g_p2p_lat = "2ms";
    //std::string g_p2p_rate = "500kb/s"; // P2P link
    //std::string g_csma_lat = "2ms";
    //std::string g_csma_rate = "500kb/s"; // P2P link
    std::string g_wifiPhyMode("DsssRate1Mbps");
    InternetStackHelper g_internetStkHlpr;
    Ipv4AddressHelper g_ipv4Hlpr;

    ////////////////////////////////////////////////////////////////////////////////

    YansWifiPhyHelper wifiPhyHlpr = YansWifiPhyHelper::Default();
    YansWifiChannelHelper wifiChnlHlpr = YansWifiChannelHelper::Default();
    Ptr<YansWifiChannel> chanel = wifiChnlHlpr.Create();

    wifiPhyHlpr.SetChannel(chanel);

    WifiHelper wifiHlpr;
    if (verbose)
    {
        wifiHlpr.EnableLogComponents(); // Turn on all Wifi logging
    }
    wifiHlpr.SetStandard(WIFI_PHY_STANDARD_80211n_2_4GHZ); // i have changed it to 802.11n2.4GHz

    // Add a non-QoS upper mac, and disable rate control
    NqosWifiMacHelper wifiMacHlpr = NqosWifiMacHelper::Default();
    wifiHlpr.SetRemoteStationManager("ns3::ConstantRateWifiManager",
            "DataMode", StringValue(g_wifiPhyMode),
            "ControlMode", StringValue(g_wifiPhyMode));
    // Set it to adhoc mode
    wifiMacHlpr.SetType("ns3::AdhocWifiMac");
    ////////////////////////////////////////////////////////////////////
    
    
    Ptr<AP_STA_Cluster> apStaCluster = CreateObject<AP_STA_Cluster> ();
    apStaCluster->setup(100, 1, wifiHlpr, wifiPhyHlpr, wifiMacHlpr, g_ipv4Hlpr, g_internetStkHlpr);
    apStaCluster->startTCPfromStaToAP(0, 1000, 1);
    
    
    
    /*

    NodeContainer apCsmaNodeContainer;
    apCsmaNodeContainer.Create(m_numberOfNodeINcsmaAPcluster);


    CsmaHelper csmaHlpr;
    csmaHlpr.SetChannelAttribute("DataRate", StringValue("100Mbps")); // TODO:: change data rate
    csmaHlpr.SetChannelAttribute("Delay", TimeValue(NanoSeconds(6560))); // TODO: change delay time

    NetDeviceContainer csmaNetDevCntnr;
    csmaNetDevCntnr = csmaHlpr.Install(apCsmaNodeContainer); // all the APs will be connected in a CSMA network


    NodeContainer wifiApNodeContainer;
    wifiApNodeContainer.Add(apCsmaNodeContainer.Get(1));
    wifiApNodeContainer.Add(apCsmaNodeContainer.Get(2));

    NodeContainer wifiStaNodeContainer;
    wifiStaNodeContainer.Create(m_wifi_stationCount);

    NetDeviceContainer apDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, wifiApNodeContainer);
    NetDeviceContainer staDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, wifiStaNodeContainer);

    //Mobility  model start
    MobilityHelper mobility;

    mobility.SetPositionAllocator("ns3::GridPositionAllocator",
            "MinX", DoubleValue(0.0),
            "MinY", DoubleValue(0.0),
            "DeltaX", DoubleValue(5.0),
            "DeltaY", DoubleValue(10.0),
            "GridWidth", UintegerValue(3),
            "LayoutType", StringValue("RowFirst"));

    
    //mobility.SetMobilityModel("ns3::RandomWalk2dMobilityModel",
    //        "Bounds", RectangleValue(Rectangle(-50, 50, -50, 50)));
    //mobility.Install(wifiStaNodes);
    //

    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(wifiApNodeContainer);
    mobility.Install(wifiStaNodeContainer);
    /////Mobility model end


    g_internetStkHlpr.Install(apCsmaNodeContainer);
    g_internetStkHlpr.Install(wifiStaNodeContainer);



    Ipv4AddressHelper ipv4Hlpr;

    ipv4Hlpr.SetBase("10.1.1.0", "255.255.255.0");
    Ipv4InterfaceContainer csmaIpv4IntfcCntnr = ipv4Hlpr.Assign(csmaNetDevCntnr);


    ipv4Hlpr.SetBase("10.1.2.0", "255.255.255.0");
    Ipv4InterfaceContainer apInterfaceContainer = ipv4Hlpr.Assign(apDevicesContainer);
    Ipv4InterfaceContainer staInterfaceContainer = ipv4Hlpr.Assign(staDevicesContainer);
     */

/*
    NS_LOG_UNCOND("Create Applications.");

    uint32_t TCPpacketSize = 1000; // bytes
    uint32_t TCPnumPackets = 1;

    int sinkNodeIndex = 0;
    uint16_t sinkPort = 8080;
    Address sinkAddress(InetSocketAddress(csmaIpv4IntfcCntnr.GetAddress(sinkNodeIndex), sinkPort)); // interface of sink
    PacketSinkHelper packetSinkHelper("ns3::TcpSocketFactory", InetSocketAddress(Ipv4Address::GetAny(), sinkPort));
    ApplicationContainer sinkApps = packetSinkHelper.Install(apCsmaNodeContainer.Get(sinkNodeIndex));
    sinkApps.Start(Seconds(0.));
    sinkApps.Stop(Seconds(100.));

    int sourceNodeIndex = 0;
    Ptr<Socket> ns3TcpSocket = Socket::CreateSocket(wifiStaNodeContainer.Get(sourceNodeIndex), TcpSocketFactory::GetTypeId()); //source at n0
    // Trace Congestion window
    ns3TcpSocket->TraceConnectWithoutContext("CongestionWindow", MakeCallback(&CwndChange));
 */
    //Config::ConnectWithoutContext("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&RcvBySink));
/*
    // Create TCP application at n0
    Ptr<MyApp> app = CreateObject<MyApp> ();
    app->Setup(ns3TcpSocket, sinkAddress, TCPpacketSize, TCPnumPackets, DataRate("250Kbps"));
    wifiStaNodeContainer.Get(sourceNodeIndex)->AddApplication(app);
    app->SetStartTime(Seconds(1.));
    app->SetStopTime(Seconds(100.));
    
*/


    Ipv4GlobalRoutingHelper::PopulateRoutingTables();

    wifiPhyHlpr.EnablePcap("Adhoc_4_wifi_NEW_", apStaCluster->staDevicesContainer.Get(0), true);
    apStaCluster->csmaHlpr.EnablePcapAll("Adhoc_4_csma_NEW_", true);



    Simulator::Stop(Seconds(100.0));
    Simulator::Run();
    Simulator::Destroy();

    return 0;


}