#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/mobility-module.h"
#include "ns3/config-store-module.h"
#include "ns3/wifi-module.h"
#include "ns3/internet-module.h"
#include <fstream>
#include <string>
#include "ns3/point-to-point-module.h"
#include "ns3/applications-module.h"
#include "ns3/flow-monitor-module.h"
#include "ns3/ipv4-global-routing-helper.h" 

#include <iostream>
#include <fstream>
#include <vector>
#include <string>

using namespace ns3;

/////////////////////////////////////////////////////////////////////////////////////////////

class MyApp : public Application
{
public:

    MyApp();
    virtual ~MyApp();

    void Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate);
    void ChangeRate(DataRate newrate);

private:
    virtual void StartApplication(void);
    virtual void StopApplication(void);

    void ScheduleTx(void);
    void SendPacket(void);

    Ptr<Socket> m_socket;
    Address m_peer;
    uint32_t m_packetSize;
    uint32_t m_nPackets;
    DataRate m_dataRate;
    EventId m_sendEvent;
    bool m_running;
    uint32_t m_packetsSent;
};

MyApp::MyApp()
: m_socket(0),
m_peer(),
m_packetSize(0),
m_nPackets(0),
m_dataRate(0),
m_sendEvent(),
m_running(false),
m_packetsSent(0)
{
}

MyApp::~MyApp()
{
    m_socket = 0;
}

void
MyApp::Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate)
{
    m_socket = socket;
    m_peer = address;
    m_packetSize = packetSize;
    m_nPackets = nPackets;
    m_dataRate = dataRate;
}

void
MyApp::StartApplication(void)
{
    m_running = true;
    m_packetsSent = 0;
    m_socket->Bind();
    m_socket->Connect(m_peer);
    SendPacket();
}

void
MyApp::StopApplication(void)
{
    m_running = false;

    if (m_sendEvent.IsRunning())
    {
        Simulator::Cancel(m_sendEvent);
    }

    if (m_socket)
    {
        m_socket->Close();
    }
}

void
MyApp::SendPacket(void)
{
    Ptr<Packet> packet = Create<Packet> (m_packetSize);
    m_socket->Send(packet);

    if (++m_packetsSent < m_nPackets)
    {
        ScheduleTx();
    }
}

void
MyApp::ScheduleTx(void)
{
    if (m_running)
    {
        Time tNext(Seconds(m_packetSize * 8 / static_cast<double> (m_dataRate.GetBitRate())));
        m_sendEvent = Simulator::Schedule(tNext, &MyApp::SendPacket, this);
    }
}

void
MyApp::ChangeRate(DataRate newrate)
{
    m_dataRate = newrate;
    return;
}
/////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[])
{
    //bool verbose = false;

    CommandLine cmd;
    cmd.Parse(argc, argv);

    NodeContainer wifiNodeContainer;
    wifiNodeContainer.Create(2);

    WifiHelper wifiHlpr;
    wifiHlpr.SetStandard(WIFI_PHY_STANDARD_80211a);

    // Set to a non-QoS upper mac
    NqosWifiMacHelper wifiMacHlpr = NqosWifiMacHelper::Default();

    // Set it to adhoc mode
    wifiMacHlpr.SetType("ns3::AdhocWifiMac");

    // Set Wi-Fi rate manager
    std::string phyMode("OfdmRate54Mbps");
    wifiHlpr.SetRemoteStationManager("ns3::ConstantRateWifiManager",
            "DataMode", StringValue(phyMode),
            "ControlMode", StringValue(phyMode));

    YansWifiPhyHelper wifiPhyHlpr = YansWifiPhyHelper::Default();

#if 1
    YansWifiChannelHelper wifiChnlHlpr = YansWifiChannelHelper::Default();
#else
    YansWifiChannelHelper wifiChnlHlpr;
    wifiChnlHlpr.SetPropagationDelay("ns3::ConstantSpeedPropagationDelayModel");
    wifiChnlHlpr.AddPropagationLoss("ns3::LogDistancePropagationLossModel",
            "Exponent", DoubleValue(3.0));
#endif
    wifiPhyHlpr.SetChannel(wifiChnlHlpr.Create());
    NetDeviceContainer wifiDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, wifiNodeContainer);

    // Enable PCAP
    //wifiPhy.EnablePcap("wifi-adhoc", devices);

#if 1
    MobilityHelper mobility;

    mobility.SetPositionAllocator("ns3::GridPositionAllocator",
            "MinX", DoubleValue(0.0),
            "MinY", DoubleValue(0.0),
            "DeltaX", DoubleValue(20.0),
            "DeltaY", DoubleValue(20.0),
            "GridWidth", UintegerValue(3),
            "LayoutType", StringValue("RowFirst"));

    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(wifiNodeContainer);
#else 
    // Configure mobility
    MobilityHelper mobility;
    Ptr<ListPositionAllocator> posAlloc = CreateObject<ListPositionAllocator> ();
    posAlloc->Add(Vector(0.0, 0.0, 0.0));
    posAlloc->Add(Vector(5.0, 0.0, 0.0));
    mobility.SetPositionAllocator(posAlloc);
    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(wifiNodeContainer);
#endif
    /////////////////////////////////

    InternetStackHelper g_internetStkHlpr;
    g_internetStkHlpr.Install(wifiNodeContainer);

    Ipv4AddressHelper ipv4Hlpr;

    ipv4Hlpr.SetBase("10.1.1.0", "255.255.255.0");
    Ipv4InterfaceContainer wifiInterfaceContainer = ipv4Hlpr.Assign(wifiDevicesContainer);





    double simulationTimeSec = 3;
    uint32_t UDPpacketSize = 1500; // bytes
    uint32_t UDPnumPackets = 9999999;
    int senderID = 0;
    int receiverID = 1;


    uint16_t sinkPort = 1010;
    Address staSinkAddress;
    staSinkAddress = InetSocketAddress(wifiInterfaceContainer.GetAddress(receiverID), sinkPort); // interface of sink
    PacketSinkHelper packetSinkHelper("ns3::UdpSocketFactory", InetSocketAddress(Ipv4Address::GetAny(), sinkPort));
    ApplicationContainer sinkApps = packetSinkHelper.Install(wifiNodeContainer.Get(receiverID));
    sinkApps.Start(Seconds(0.0));
    sinkApps.Stop(Seconds(simulationTimeSec));

    Ptr<Socket> ns3UdpSocket = Socket::CreateSocket(wifiNodeContainer.Get(senderID), UdpSocketFactory::GetTypeId());

    Ptr<MyApp> app = CreateObject<MyApp> ();
    app->Setup(ns3UdpSocket, staSinkAddress, UDPpacketSize, UDPnumPackets, DataRate("54Mbps"));
    wifiNodeContainer.Get(senderID)->AddApplication(app);
    //Config::ConnectWithoutContext("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&RcvBySink));
    app->SetStartTime(Seconds(0.0));
    app->SetStopTime(Seconds(simulationTimeSec));

    Simulator::Stop(Seconds(simulationTimeSec));
    Simulator::Run();
    Simulator::Destroy();


    uint32_t totalUDPpacketsThrough = DynamicCast<PacketSink> (sinkApps.Get(0))->GetTotalRx();
    double UDPthroughput = totalUDPpacketsThrough * 8 / (simulationTimeSec * 1000000.0); //Mbit/s
    //std::cout << totalUDPpacketsThrough << " Mbps\n";
    std::cout << "total UDP byte : " << totalUDPpacketsThrough << "; UDP throughput in AP :  " << UDPthroughput << " Mbps\n";

    return 0;

}
