/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ns3/core-module.h"
#include "ns3/point-to-point-module.h"
#include "ns3/network-module.h"
#include "ns3/applications-module.h"
#include "ns3/wifi-module.h"
#include "ns3/mobility-module.h"
#include "ns3/csma-module.h"
#include "ns3/internet-module.h"

// Default Network Topology
//
//   Wifi 10.1.3.0
//                 AP
//  *    *    *    *
//  |    |    |    |    10.1.1.0
// n5   n6   n7   n0 -------------- n1
//                   point-to-point  

using namespace ns3;

NS_LOG_COMPONENT_DEFINE("ThirdScriptExample");

void ReceivePacket(Ptr<Socket> socket)
{
    while (socket->Recv())
    {
        NS_LOG_UNCOND("Received one packet!");
    }
}

static void GenerateTraffic(Ptr<Socket> socket, uint32_t pktSize,
        uint32_t pktCount, Time pktInterval)
{
    if (pktCount > 0)
    {
        socket->Send(Create<Packet> (pktSize));
        Simulator::Schedule(pktInterval, &GenerateTraffic,
                socket, pktSize, pktCount - 1, pktInterval);
    }
    else
    {
        socket->Close();
    }
}

int
main(int argc, char *argv[])
{
    bool verbose = true;
    uint32_t nCsma = 3;
    uint32_t nWifi = 3;
    
    
    CommandLine cmd;
    cmd.AddValue("nCsma", "Number of \"extra\" CSMA nodes/devices", nCsma);
    cmd.AddValue("nWifi", "Number of wifi STA devices", nWifi);
    cmd.AddValue("verbose", "Tell echo applications to log if true", verbose);

    cmd.Parse(argc, argv);

    if (nWifi > 18)
    {
        std::cout << "Number of wifi nodes " << nWifi <<
                " specified exceeds the mobility bounding box" << std::endl;
        exit(1);
    }

    if (verbose)
    {
        LogComponentEnable("UdpEchoClientApplication", LOG_LEVEL_INFO);
        LogComponentEnable("UdpEchoServerApplication", LOG_LEVEL_INFO);
    }

    NodeContainer p2pNodeContainer;
    p2pNodeContainer.Create(2);

    PointToPointHelper pointToPointHlpr;
    pointToPointHlpr.SetDeviceAttribute("DataRate", StringValue("5Mbps"));
    pointToPointHlpr.SetChannelAttribute("Delay", StringValue("2ms"));

    NetDeviceContainer p2pNetDevCntnr;
    p2pNetDevCntnr = pointToPointHlpr.Install(p2pNodeContainer);


    
    NodeContainer wifiStaNodeCtnr;
    wifiStaNodeCtnr.Create(nWifi);
    NodeContainer wifiApNodeCtnr = p2pNodeContainer.Get(0);

    YansWifiChannelHelper channelHlpr = YansWifiChannelHelper::Default();
    YansWifiPhyHelper phyHlpr = YansWifiPhyHelper::Default();
    phyHlpr.SetChannel(channelHlpr.Create());

    WifiHelper wifi = WifiHelper::Default();
    wifi.SetRemoteStationManager("ns3::AarfWifiManager");

    NqosWifiMacHelper mac = NqosWifiMacHelper::Default();

    Ssid ssid = Ssid("ns-3-ssid");
    mac.SetType("ns3::StaWifiMac",
            "Ssid", SsidValue(ssid),
            "ActiveProbing", BooleanValue(false));

    NetDeviceContainer staDevices;
    staDevices = wifi.Install(phyHlpr, mac, wifiStaNodeCtnr);

    mac.SetType("ns3::ApWifiMac",
            "Ssid", SsidValue(ssid));

    NetDeviceContainer apDevices;
    apDevices = wifi.Install(phyHlpr, mac, wifiApNodeCtnr);

    MobilityHelper mobility;

    mobility.SetPositionAllocator("ns3::GridPositionAllocator",
            "MinX", DoubleValue(0.0),
            "MinY", DoubleValue(0.0),
            "DeltaX", DoubleValue(5.0),
            "DeltaY", DoubleValue(10.0),
            "GridWidth", UintegerValue(3),
            "LayoutType", StringValue("RowFirst"));

    mobility.SetMobilityModel("ns3::RandomWalk2dMobilityModel",
            "Bounds", RectangleValue(Rectangle(-50, 50, -50, 50)));
    mobility.Install(wifiStaNodeCtnr);

    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(wifiApNodeCtnr);

    InternetStackHelper stack;

    stack.Install(p2pNodeContainer.Get(1));
    stack.Install(wifiApNodeCtnr);
    stack.Install(wifiStaNodeCtnr);

    Ipv4AddressHelper address;

    address.SetBase("10.1.1.0", "255.255.255.0");
    Ipv4InterfaceContainer p2pInterfaces;
    p2pInterfaces = address.Assign(p2pNetDevCntnr);


    address.SetBase("10.1.2.0", "255.255.255.0");
    Ipv4InterfaceContainer stationInterfaces = address.Assign(staDevices);
    Ipv4InterfaceContainer apInterfaces = address.Assign(apDevices);

 ///////////////////////// add application ////////////////////////////////////////////
#if 1
    
    uint32_t packetSize = 1000; // bytes
    uint32_t numPackets = 1;
    double interval = 1.0; // seconds
    Time interPacketInterval = Seconds(interval);
     
    TypeId tid = TypeId::LookupByName("ns3::UdpSocketFactory");
    Ptr<Socket> recvSink = Socket::CreateSocket(p2pNodeContainer.Get(1), tid);
    InetSocketAddress local = InetSocketAddress(Ipv4Address::GetAny(), 80);
    recvSink->Bind(local);
    recvSink->SetRecvCallback(MakeCallback(&ReceivePacket));

    //Ptr<Socket> source = Socket::CreateSocket(wifiStaNodeContainer.Get(0), tid);
    Ptr<Socket> source = Socket::CreateSocket(wifiStaNodeCtnr.Get(nWifi - 1), tid);
    //InetSocketAddress remote = InetSocketAddress(Ipv4Address("255.255.255.255"), 80);
    InetSocketAddress remote = InetSocketAddress(stationInterfaces.GetAddress(nWifi - 1) ,80);
    //source->SetAllowBroadcast(true);
    source->Connect(remote);
    
    Simulator::ScheduleWithContext(source->GetNode()->GetId(),
            Seconds(1.0), &GenerateTraffic,
            source, packetSize, numPackets, interPacketInterval);
    
#else    
    UdpEchoServerHelper echoServer(9);
    //ApplicationContainer serverApps = echoServer.Install(csmaNodeContainer.Get(nCsma));
    ApplicationContainer serverApps = echoServer.Install(p2pNodeContainer.Get(1));
    serverApps.Start(Seconds(1.0));
    serverApps.Stop(Seconds(10.0));


    //UdpEchoClientHelper echoClient(p2pInterfaces.GetAddress(1), 9);
    UdpEchoClientHelper echoClient(p2pInterfaces.GetAddress(1), 9);
    echoClient.SetAttribute("MaxPackets", UintegerValue(1));
    echoClient.SetAttribute("Interval", TimeValue(Seconds(1.0)));
    echoClient.SetAttribute("PacketSize", UintegerValue(1024));

    ApplicationContainer clientApps =
            echoClient.Install(wifiStaNodeCtnr.Get(nWifi - 1));
    clientApps.Start(Seconds(2.0));
    clientApps.Stop(Seconds(10.0));
#endif 
 ///////////////////////// add application ////////////////////////////////////////////
    
    Ipv4GlobalRoutingHelper::PopulateRoutingTables();

    Simulator::Stop(Seconds(10.0));

    pointToPointHlpr.EnablePcapAll("third");
    phyHlpr.EnablePcap("third", apDevices.Get(0));

    Simulator::Run();
    Simulator::Destroy();
    return 0;
}
