/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2005,2006 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Mathieu Lacage <mathieu.lacage@sophia.inria.fr>
 */

#ifndef SIMULATOR_IMPL_H
#define SIMULATOR_IMPL_H

#include "event-impl.h"
#include "event-id.h"
#include "nstime.h"
#include "object.h"
#include "object-factory.h"
#include "ptr.h"
#include "ns3/address.h"
#include "ns3/ipv4-address.h"
#include "ns3/mac48-address.h"
#include "ns3/log-macros-enabled.h"
#include <map>
#include <stdio.h>

#include "ns3/wifi-net-device.h"


//#define SBA_CHNL_SNS_DURATION_IN_MILLI_SEC  1000
#define SBA_DEFAULT_CHNL_SNS_DURATION_IN_MILLI_SEC  1000
#define SBA_DEFAUT_CHNL_CHANGE_INTERVAL_IN_MILLI_SEC  1000

/**
 * \file
 * \ingroup simulator
 * Declaration of class ns3::SimulatorImpl.
 */

namespace ns3
{

    class Scheduler;

    /**
     * \ingroup simulator
     *
     * The SimulatorImpl base class.
     *
     * \todo Define what the simulation or event context means.
     */

    enum SIMULATION_TYPE_Plaban
    {
        NORMAL_SIMULATION,
        SIMULATION_MULTI_RADIO_SINGLE_CHANNEL = 20,
        SIMULATION_MULTI_RADIO_MULTI_CHANNEL = 30,
    };

    enum Addr_TYPE_sba
    {
        WIFI_ADDR = 10,
        CSMA_ADDR = 11,
    };

    enum NodeType_Plaban
    {
        AGGREGATOR = 10,
        AP_MODULE = 11,
        NORMAL_NODE = 12,
    };

    enum CHANNEL_Plaban
    {
        CHANNEL_ONE = 1,
        CHANNEL_TWO = 2,
        CHANNEL_THREE = 3,
        CHANNEL_DUMMY_FOR_THE_AGGREGATOR = 4, // Aggregator is a CSMA node, so it has no wifi channel
    };

    /*
     *....................AGGREGATOR........................
     *...........//...........||............\\.............. CSMA NETWORK
     *......AP_MODULE.....AP_MODULE......AP_MODULE..........
     * ^^^^channel 1^^^^^^^channel 2^^^^^ channel 3^^^^^^^^^ WIFI NETWORK
     * ()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()
     * NORMAL_NODE 1        NORMAL_NODE 2       ....     NORMAL_NODE N
     * 
     */


    class Statistics
    {
    public:
        uint32_t nodeID;
        CHANNEL_Plaban channel;
        int numberOfPktDrop;
        int numberOfReTransmisison;
        int numberOfPktRcv;
        double RxPwr;
    };

    class Sta_AP_statistics
    {
    public:

        Sta_AP_statistics()
        {
            this->perChnlNodeStatMAP.clear();
        }

        uint32_t groupID;
        std::map<CHANNEL_Plaban, Statistics> perChnlNodeStatMAP;
    };

    class ChannelApID_pair
    {
    public:
        uint32_t apID;
        CHANNEL_Plaban channel;

        ChannelApID_pair(uint32_t _apID, CHANNEL_Plaban _channel)
        {
            this->apID = _apID;
            this->channel = _channel;
        }
    };

    class NodeGroupChnlTypeInfo
    {
    public:
        int group;
        CHANNEL_Plaban chnl;
        Time channelLastSetTimeMilliS;
        NodeType_Plaban type;
        bool isInitialized;

        NodeGroupChnlTypeInfo(int _group, CHANNEL_Plaban _chnl, NodeType_Plaban _type)
        {
            this->chnl = _chnl;
            this->group = _group;
            this->type = _type;
            this->isInitialized = true;
            this->channelLastSetTimeMilliS = MilliSeconds(0.0);
        }

        NodeGroupChnlTypeInfo()
        {
            this->isInitialized = false;
        }

    };

    class PktDropCounter
    {
    public:

        //double chnlSenseDurationInSec;
        Time chnlSenseDuration;
        std::map<Time, int> sba_time_pktDrop_MAP; // this second int (pktDrop) is just a dummy int. We only need the total number within the specific time.

        PktDropCounter(Time _channelSnsDuration)
        {
            this->chnlSenseDuration = _channelSnsDuration;
        }

        void increaseDroppedPkt(Time _currentTime)
        {
            std::pair < std::map<Time, int>::iterator, bool> ret;
            ret = this->sba_time_pktDrop_MAP.insert(std::pair<Time, int> (_currentTime, 1234)); // note: this 1234 is just a dummy integer. we only need the number of entry added in this map.
        }

        int getPktDropCount(Time _currentTime)
        {
            //double earliestTime = _currentTimeInSec - this->chnlSenseDurationInSec;
            Time earliestTime = _currentTime - this->chnlSenseDuration;

            std::map<Time, int>::iterator it;

            for (it = this->sba_time_pktDrop_MAP.begin(); it != this->sba_time_pktDrop_MAP.end(); ++it)
            {
                //if (it->first < earliestTime)
                if (-1 == (it->first.Compare(earliestTime)))
                    this->sba_time_pktDrop_MAP.erase(it);
                else
                    break;
            }

            return this->sba_time_pktDrop_MAP.size();
        }
    };

    class PktReTxCounter
    {
    public:

        //double chnlSenseDurationInSec;
        Time chnlSenseDuration;
        std::map<Time, int> sba_time_pktReTx_MAP; // this second int (pktDrop) is just a dummy int. We only need the total number within the specific time.

        PktReTxCounter(Time _channelSnsDuration)
        {
            this->chnlSenseDuration = _channelSnsDuration;
        }

        void increaseReTxPkt(Time _currentTime)
        {
            std::pair < std::map<Time, int>::iterator, bool> ret;
            ret = this->sba_time_pktReTx_MAP.insert(std::pair<Time, int> (_currentTime, 1234)); // note: this 1234 is just a dummy integer. we only need the number of entry added in this map.
        }

        int getReTxCount(Time _currentTime)
        {
            //double earliestTime = _currentTimeInSec - this->chnlSenseDurationInSec;
            Time earliestTime = _currentTime - this->chnlSenseDuration;

            std::map<Time, int>::iterator it;

            for (it = this->sba_time_pktReTx_MAP.begin(); it != this->sba_time_pktReTx_MAP.end(); ++it)
            {
                //if (it->first < earliestTime)
                if (-1 == (it->first.Compare(earliestTime)))
                    this->sba_time_pktReTx_MAP.erase(it);
                else
                    break;
            }

            return this->sba_time_pktReTx_MAP.size();
        }
    };

    class RcvPktInfoCounter
    {
    public:

        class RcvdPktInfo
        {
        public:
            double m_rxPowerW;
            //double rx_duration;
        };

        //double chnlSenseDurationInSec;
        Time chnlSenseDuration;
        std::map<Time, RcvdPktInfo> sba_time_pktRcvInfo_MAP;
        double totalRcvPwr;

        RcvPktInfoCounter(Time _channelSnsDuration)
        {
            this->totalRcvPwr = 0.0;
            this->chnlSenseDuration = _channelSnsDuration;
        }

        //void addRcvdPkt(double _currentTimeInSec, double rxPwrW, double rxDuration)

        void addRcvdPkt(Time _currentTime, double rxPwrW, Time rxDuration)
        {
            std::pair < std::map<Time, RcvdPktInfo>::iterator, bool> ret;
            RcvdPktInfo rPktInfo;
            rPktInfo.m_rxPowerW = rxPwrW;
            //rPktInfo.rx_duration = rxDuration;

            Time rxEndTime = _currentTime + rxDuration;

            this->totalRcvPwr += rPktInfo.m_rxPowerW;

            ret = this->sba_time_pktRcvInfo_MAP.insert(std::pair<Time, RcvdPktInfo> (rxEndTime, rPktInfo)); // note: this 1234 is just a dummy integer. we only need the number of entry added in this map.
        }

        double getTotalRxPwrCount(Time _currentTime, int & OUTPUT_receivedPktCount)
        {
            Time earliestTime = _currentTime - this->chnlSenseDuration;
            std::map<Time, RcvdPktInfo>::iterator it;

            for (it = this->sba_time_pktRcvInfo_MAP.begin(); it != this->sba_time_pktRcvInfo_MAP.end(); ++it)
            {
                if (-1 == (it->first.Compare(earliestTime)))
                {
                    this->totalRcvPwr -= it->second.m_rxPowerW;
                    this->sba_time_pktRcvInfo_MAP.erase(it);
                }
                else
                    break;
            }

            OUTPUT_receivedPktCount = this->sba_time_pktRcvInfo_MAP.size();

            return this->totalRcvPwr;
        }

    };

    class NodeID_AdrType
    {
    public:
        uint32_t nodeID;
        Addr_TYPE_sba adrType;

        NodeID_AdrType(uint32_t _nodeID, Addr_TYPE_sba _adrType)
        {
            this->nodeID = _nodeID;
            this->adrType = _adrType;
        }
    };

    class SimulatorImpl : public Object
    {
    public:

        int sba_totalRetransmissionCount;

        void sba_increaseTotalRetransmissionCount(int _nodeID)
        {
            this->sba_totalRetransmissionCount++;
        }

        int sba_getTotalRetransmissionCount()
        {
            return this->sba_totalRetransmissionCount;
        }

        int sba_totalPktDropCount;
        
        void sba_increaseTotalPktDropCount(int _nodeID)
        {
            this->sba_totalPktDropCount++;
        }
        
        int sba_getTotalPktDropCount()
        {
            return this->sba_totalPktDropCount;
        }

        std::map<NodeType_Plaban, Sta_AP_statistics> sba_getSta_AP_statisticsMap(Time _currentTime, uint32_t _nodeID)
        {
            std::map<NodeType_Plaban, Sta_AP_statistics> mapToBeReturned;

            /////////////////////////////////////////////////// prepare stat for Normal Node //////////////////////////////////////////////////////////////////

            NodeGroupChnlTypeInfo stationNodeInfo = this->sba_getNodeInfo_GrpID_currentChnl_nodeType(_nodeID);
            Sta_AP_statistics stationStat;
            stationStat.groupID = stationNodeInfo.group;

            NS_ASSERT(NORMAL_NODE == stationNodeInfo.type); // this will be called on behalf of a station

            Statistics nodeStat;
            nodeStat.nodeID = _nodeID;
            nodeStat.channel = stationNodeInfo.chnl;
            nodeStat.RxPwr = this->sba_getPktRcvPwrCount(_currentTime, nodeStat.nodeID, nodeStat.channel, nodeStat.numberOfPktRcv);
            nodeStat.numberOfPktDrop = this->sba_getPktDropCount_onCurrentChnl(_currentTime, nodeStat.nodeID);
            nodeStat.numberOfReTransmisison = this->sba_getPktReTxCount_onCurrentChnl(_currentTime, nodeStat.nodeID);


            std::pair < std::map<CHANNEL_Plaban, Statistics>::iterator, bool> ret;
            ret = stationStat.perChnlNodeStatMAP.insert(std::pair<CHANNEL_Plaban, Statistics>(nodeStat.channel, nodeStat));

            std::list<CHANNEL_Plaban> remainingChannelList;
            remainingChannelList.push_back(CHANNEL_ONE);
            remainingChannelList.push_back(CHANNEL_TWO);
            remainingChannelList.push_back(CHANNEL_THREE);


            remainingChannelList.remove(nodeStat.channel);

            std::list<CHANNEL_Plaban>::iterator itList;
            for (itList = remainingChannelList.begin(); itList != remainingChannelList.end(); ++itList)
            {
                //CHANNEL_Plaban channel = *itList;
                Statistics tempNodeStat;
                tempNodeStat.nodeID = _nodeID;
                tempNodeStat.channel = *itList;
                tempNodeStat.RxPwr = this->sba_getPktRcvPwrCount(_currentTime, tempNodeStat.nodeID, tempNodeStat.channel, tempNodeStat.numberOfPktRcv);
                tempNodeStat.numberOfPktDrop = -6532;
                tempNodeStat.numberOfReTransmisison = -6532;

                ret = stationStat.perChnlNodeStatMAP.insert(std::pair<CHANNEL_Plaban, Statistics>(tempNodeStat.channel, tempNodeStat));
            }

            std::pair < std::map<NodeType_Plaban, Sta_AP_statistics>::iterator, bool > ret2;
            ret2 = mapToBeReturned.insert(std::pair<NodeType_Plaban, Sta_AP_statistics>(NORMAL_NODE, stationStat));
            /////////////////////////////////////////////////// prepare stat for AP //////////////////////////////////////////////////////////////////
            Sta_AP_statistics apStat;
            apStat.groupID = stationStat.groupID; // they will be in the same channel


            std::list<ChannelApID_pair> chnlAPlist = this->sba_get_group_apNodeList(apStat.groupID);
            std::list<ChannelApID_pair>::iterator it_chnlAPList;

            for (it_chnlAPList = chnlAPlist.begin(); it_chnlAPList != chnlAPlist.end(); ++it_chnlAPList)
            {
                ChannelApID_pair chnl_ap_pair = *it_chnlAPList;

                Statistics apSingleNodeStat;
                apSingleNodeStat.nodeID = chnl_ap_pair.apID;
                apSingleNodeStat.channel = chnl_ap_pair.channel;

                apSingleNodeStat.RxPwr = this->sba_getPktRcvPwrCount(_currentTime, apSingleNodeStat.nodeID, apSingleNodeStat.channel, apSingleNodeStat.numberOfPktRcv);
                apSingleNodeStat.numberOfPktDrop = this->sba_getPktDropCount_onCurrentChnl(_currentTime, apSingleNodeStat.nodeID);
                apSingleNodeStat.numberOfReTransmisison = this->sba_getPktReTxCount_onCurrentChnl(_currentTime, apSingleNodeStat.nodeID);

                ret = apStat.perChnlNodeStatMAP.insert(std::pair<CHANNEL_Plaban, Statistics>(apSingleNodeStat.channel, apSingleNodeStat));
            }

            ret2 = mapToBeReturned.insert(std::pair<NodeType_Plaban, Sta_AP_statistics>(AP_MODULE, apStat));


            return mapToBeReturned;
        }

        SimulatorImpl()
        {
            this->sba_channelSnsDuration_millisecond = MilliSeconds(SBA_DEFAULT_CHNL_SNS_DURATION_IN_MILLI_SEC);
            this->sba_defaultChnlChangeInterval_inMilliSec = MilliSeconds(SBA_DEFAUT_CHNL_CHANGE_INTERVAL_IN_MILLI_SEC);
            this->sba_totalPktDropCount = 0;
            this->sba_totalRetransmissionCount = 0;
        }

        SIMULATION_TYPE_Plaban simulationType;

        SIMULATION_TYPE_Plaban sba_getSimulationType()
        {
            return this->simulationType;
        }

        void sba_setSimulationType(SIMULATION_TYPE_Plaban simulationType)
        {
            this->simulationType = simulationType;
        }

        int plaban_test;

        Time sba_channelSnsDuration_millisecond;

        void sba_setChannelSnsDuration_inMillisecond(int millisecond)
        {
            this->sba_channelSnsDuration_millisecond = MilliSeconds(millisecond);
        }

        Time sba_getChannelSnsDuration_inMillisecond()
        {
            return this->sba_channelSnsDuration_millisecond;
        }

        Time sba_defaultChnlChangeInterval_inMilliSec;

        void sba_setChannelChangeInterval_inMillisecond(int millisecond)
        {
            this->sba_defaultChnlChangeInterval_inMilliSec = MilliSeconds(millisecond);
        }

        Time sba_getChannelChangeInterval_inMillisecond()
        {
            return this->sba_defaultChnlChangeInterval_inMilliSec;
        }

        std::map< int, std::map< CHANNEL_Plaban, PktDropCounter> > sba_packetDropCounterPerNodePerChnl_MAP;
        std::map< int, std::map< CHANNEL_Plaban, PktReTxCounter> > sba_packetReTXcounterPerNodePerChnl_MAP;

        std::map< int, std::map< CHANNEL_Plaban, PktDropCounter> > sba_get_sba_packetDropCounterPerNodePerChnl_MAP()
        {
            return this->sba_packetDropCounterPerNodePerChnl_MAP;
        }

        std::map< int, std::map< CHANNEL_Plaban, PktReTxCounter> > sba_get_sba_packetReTxCounterPerNodePerChnl_MAP()
        {
            return this->sba_packetReTXcounterPerNodePerChnl_MAP;
        }

        std::map< int, std::map< CHANNEL_Plaban, RcvPktInfoCounter> > sba_pktRcvInfoCounterPerNodePerChnl_MAP;

        std::map< int, std::map< CHANNEL_Plaban, RcvPktInfoCounter> > sba_get_sba_pktRcvInfoCounterPerNodePerChnl_MAP()
        {
            return this->sba_pktRcvInfoCounterPerNodePerChnl_MAP;
        }

        std::map<Ipv4Address, NodeID_AdrType > sba_IP_nodeID_adrType_MAP;

        std::map<Ipv4Address, NodeID_AdrType > sba_get_IP_nodeID_adrTypeMap()
        {
            return this->sba_IP_nodeID_adrType_MAP;
        }

        std::map< uint32_t, std::map<Addr_TYPE_sba, Mac48Address> > sba_nodeID_macType_MacAdr_MAP;

        std::map<int, NodeGroupChnlTypeInfo> sba_node_group_channel_map;
        std::map<int, std::list<ChannelApID_pair> > sba_groupID_ChannelApIDPair_Map;

        std::map< int, std::map<CHANNEL_Plaban, Mac48Address> > sba_csmaGroupBasedMacChnlMap;

        std::map< int, std::map<CHANNEL_Plaban, Mac48Address> > sba_get_group_chnl_mac_CSMA_MAPPING()
        {
            return this->sba_csmaGroupBasedMacChnlMap;
        }

        std::map< int, std::map<CHANNEL_Plaban, Mac48Address> > sba_wifiAP_GroupBasedMacChnlMap;

        std::map< int, std::map<CHANNEL_Plaban, Mac48Address> > sba_get_group_chnl_mac_WIFI_MAPPING()
        {
            return this->sba_wifiAP_GroupBasedMacChnlMap;
        }

        std::map< int, std::map<CHANNEL_Plaban, Ptr<NetDevice> > > sba_wifiAP_GroupBasedChnl_netDev_Map;


        std::map< int, std::map<CHANNEL_Plaban, Ipv4Address> > sba_csmaGroupBased_IPAdr_ChnlMap;

        std::map< int, std::map<CHANNEL_Plaban, Ipv4Address> > sba_get_csmaGroupBased_IPAdr_ChnlMap()
        {
            return this->sba_csmaGroupBased_IPAdr_ChnlMap;
        }

        std::map< int, std::map<CHANNEL_Plaban, Ipv4Address> > sba_wifiGroupBased_IPAdr_ChnlMap;

        std::map< int, std::map<CHANNEL_Plaban, Ipv4Address> > sba_get_wifiGroupBased_IPAdr_ChnlMap()
        {
            return this->sba_wifiGroupBased_IPAdr_ChnlMap;
        }

        std::map< uint32_t, std::map<Addr_TYPE_sba, Mac48Address> > get_sba_nodeID_macType_MacAdr_MAP()
        {
            return this->sba_nodeID_macType_MacAdr_MAP;
        }

        double sba_getPktRcvPwrCount(Time _currentTime, int _nodeID, CHANNEL_Plaban channel, int & OUTPUT_receivedPktCount)
        {

            std::map< int, std::map< CHANNEL_Plaban, RcvPktInfoCounter> >::iterator it;
            //NodeGroupChnlTypeInfo ngctInfo = this->sba_getNodeInfo_GrpID_currentChnl_nodeType(_nodeID);
            //CHANNEL_Plaban channel = ngctInfo.chnl;

            it = this->sba_pktRcvInfoCounterPerNodePerChnl_MAP.find(_nodeID);

            if (this->sba_pktRcvInfoCounterPerNodePerChnl_MAP.end() == it)
            {
                OUTPUT_receivedPktCount = 0;
                return 0.0; // no entry found
            }
            else
            {
                std::map< CHANNEL_Plaban, RcvPktInfoCounter>::iterator it2;

                it2 = it->second.find(channel);
                if (it->second.end() == it2)
                {
                    OUTPUT_receivedPktCount = 0;
                    return 0.0; // no entry found
                }
                else
                {
                    return it2->second.getTotalRxPwrCount(_currentTime, OUTPUT_receivedPktCount);
                }

            }
        }

        void sba_add_rcvd_pkt_info(Time _currentTime, int _nodeID, CHANNEL_Plaban _channel, double rxPwrW, Time rxDuration)
        {
            std::map< int, std::map< CHANNEL_Plaban, RcvPktInfoCounter> >::iterator it;

            it = this->sba_pktRcvInfoCounterPerNodePerChnl_MAP.find(_nodeID);

            if (this->sba_pktRcvInfoCounterPerNodePerChnl_MAP.end() == it)
            {
                RcvPktInfoCounter pktRcvCntr(this->sba_getChannelSnsDuration_inMillisecond());
                pktRcvCntr.addRcvdPkt(_currentTime, rxPwrW, rxDuration);

                std::map<CHANNEL_Plaban, RcvPktInfoCounter> chnlPktRcvInfoMap;

                chnlPktRcvInfoMap.insert(std::pair<CHANNEL_Plaban, RcvPktInfoCounter> (_channel, pktRcvCntr));
                this->sba_pktRcvInfoCounterPerNodePerChnl_MAP.insert(std::pair< int, std::map< CHANNEL_Plaban, RcvPktInfoCounter> >(_nodeID, chnlPktRcvInfoMap));
            }
            else
            {
                std::map< CHANNEL_Plaban, RcvPktInfoCounter>::iterator it2;

                it2 = it->second.find(_channel);
                if (it->second.end() == it2)
                {
                    RcvPktInfoCounter pktRcvCntr(this->sba_getChannelSnsDuration_inMillisecond());
                    pktRcvCntr.addRcvdPkt(_currentTime, rxPwrW, rxDuration); // increaseDroppedPkt(_currentTimeInSec);

                    it->second.insert(std::pair<CHANNEL_Plaban, RcvPktInfoCounter>(_channel, pktRcvCntr));
                }
                else
                {
                    it2->second.addRcvdPkt(_currentTime, rxPwrW, rxDuration);
                }
            }

        }

        int sba_getPktDropCount_onCurrentChnl(Time _currentTime, int _nodeID)
        {
            CHANNEL_Plaban channel = this->sba_getNodeCurrentChannel(_nodeID);

            std::map< int, std::map< CHANNEL_Plaban, PktDropCounter> >::iterator it;

            it = this->sba_packetDropCounterPerNodePerChnl_MAP.find(_nodeID);

            if (this->sba_packetDropCounterPerNodePerChnl_MAP.end() == it)
            {
                return 0; // no entry found
            }
            else
            {
                std::map< CHANNEL_Plaban, PktDropCounter>::iterator it2;

                it2 = it->second.find(channel);
                if (it->second.end() == it2)
                {
                    return 0; // no entry found
                }
                else
                {
                    return it2->second.getPktDropCount(_currentTime);
                }

            }
        }

        void sba_increase_pkt_drop(Time _currentTime, int _nodeID, CHANNEL_Plaban _channel)
        {
            std::map< int, std::map< CHANNEL_Plaban, PktDropCounter> >::iterator it;

            it = this->sba_packetDropCounterPerNodePerChnl_MAP.find(_nodeID);

            if (this->sba_packetDropCounterPerNodePerChnl_MAP.end() == it)
            {
                PktDropCounter pktDrpCntr(this->sba_getChannelSnsDuration_inMillisecond());
                pktDrpCntr.increaseDroppedPkt(_currentTime);

                std::map<CHANNEL_Plaban, PktDropCounter> chnlPktdropMap;

                chnlPktdropMap.insert(std::pair<CHANNEL_Plaban, PktDropCounter> (_channel, pktDrpCntr));
                this->sba_packetDropCounterPerNodePerChnl_MAP.insert(std::pair< int, std::map< CHANNEL_Plaban, PktDropCounter> >(_nodeID, chnlPktdropMap));
            }
            else
            {
                //printf("\n\n\n hi \n\n\n");
                std::map< CHANNEL_Plaban, PktDropCounter>::iterator it2;

                it2 = it->second.find(_channel);
                if (it->second.end() == it2)
                {
                    PktDropCounter pktDrpCntr(this->sba_getChannelSnsDuration_inMillisecond());
                    pktDrpCntr.increaseDroppedPkt(_currentTime);

                    it->second.insert(std::pair<CHANNEL_Plaban, PktDropCounter>(_channel, pktDrpCntr));
                }
                else
                {
                    it2->second.increaseDroppedPkt(_currentTime);
                }
            }

        }

        int sba_getPktReTxCount_onCurrentChnl(Time _currentTime, int _nodeID)
        {
            CHANNEL_Plaban channel = this->sba_getNodeCurrentChannel(_nodeID);

            std::map< int, std::map< CHANNEL_Plaban, PktReTxCounter> >::iterator it;

            it = this->sba_packetReTXcounterPerNodePerChnl_MAP.find(_nodeID);

            if (this->sba_packetReTXcounterPerNodePerChnl_MAP.end() == it)
            {
                return 0; // no entry found
            }
            else
            {
                std::map< CHANNEL_Plaban, PktReTxCounter>::iterator it2;

                it2 = it->second.find(channel);
                if (it->second.end() == it2)
                {
                    return 0; // no entry found
                }
                else
                {
                    return it2->second.getReTxCount(_currentTime);
                }

            }
        }

        void sba_increase_ReTransmissionCount(Time _currentTime, int _nodeID, CHANNEL_Plaban _channel)
        {
            std::map< int, std::map< CHANNEL_Plaban, PktReTxCounter> >::iterator it;

            it = this->sba_packetReTXcounterPerNodePerChnl_MAP.find(_nodeID);

            if (this->sba_packetReTXcounterPerNodePerChnl_MAP.end() == it)
            {
                //std::cout <<" 314 sba_increase_ReTransmissionCount( ) for the first time for nodeID = "<<_nodeID << endl;

                PktReTxCounter pktReTxCntr(this->sba_getChannelSnsDuration_inMillisecond());
                pktReTxCntr.increaseReTxPkt(_currentTime);

                std::map<CHANNEL_Plaban, PktReTxCounter> chnlPktReTxMap;

                chnlPktReTxMap.insert(std::pair<CHANNEL_Plaban, PktReTxCounter> (_channel, pktReTxCntr));
                this->sba_packetReTXcounterPerNodePerChnl_MAP.insert(std::pair< int, std::map< CHANNEL_Plaban, PktReTxCounter> >(_nodeID, chnlPktReTxMap));
            }
            else
            {
                std::map< CHANNEL_Plaban, PktReTxCounter>::iterator it2;

                it2 = it->second.find(_channel);
                if (it->second.end() == it2)
                {
                    PktReTxCounter pktReTxCntr(this->sba_getChannelSnsDuration_inMillisecond());
                    pktReTxCntr.increaseReTxPkt(_currentTime);

                    it->second.insert(std::pair<CHANNEL_Plaban, PktReTxCounter>(_channel, pktReTxCntr));
                }
                else
                {
                    it2->second.increaseReTxPkt(_currentTime);
                }
            }

        }

        void sba_SET_CHNL_CSMA_IPAdr(int _groupID, CHANNEL_Plaban _channel, Ipv4Address _ipAdr)
        {
            std::map< int, std::map<CHANNEL_Plaban, Ipv4Address> >::iterator it;
            it = this->sba_csmaGroupBased_IPAdr_ChnlMap.find(_groupID);

            if (this->sba_csmaGroupBased_IPAdr_ChnlMap.end() == it)
            {
                std::map<CHANNEL_Plaban, Ipv4Address> channel_IP_MAP;
                channel_IP_MAP.insert(std::pair<CHANNEL_Plaban, Ipv4Address>(_channel, _ipAdr));
                this->sba_csmaGroupBased_IPAdr_ChnlMap.insert(std::pair< int, std::map<CHANNEL_Plaban, Ipv4Address> >(_groupID, channel_IP_MAP));
            }
            else
            {
                std::pair < std::map<CHANNEL_Plaban, Ipv4Address>::iterator, bool> ret;
                ret = it->second.insert(std::pair<CHANNEL_Plaban, Ipv4Address>(_channel, _ipAdr));
                NS_ASSERT(ret.second);
            }
        }

        void sba_SET_CHNL_WIFI_IPAdr(int _groupID, CHANNEL_Plaban _channel, Ipv4Address _ipAdr)
        {
            std::map< int, std::map<CHANNEL_Plaban, Ipv4Address> >::iterator it;
            it = this->sba_wifiGroupBased_IPAdr_ChnlMap.find(_groupID);

            if (this->sba_wifiGroupBased_IPAdr_ChnlMap.end() == it)
            {
                std::map<CHANNEL_Plaban, Ipv4Address> channel_IP_MAP;
                channel_IP_MAP.insert(std::pair<CHANNEL_Plaban, Ipv4Address>(_channel, _ipAdr));
                this->sba_wifiGroupBased_IPAdr_ChnlMap.insert(std::pair< int, std::map<CHANNEL_Plaban, Ipv4Address> >(_groupID, channel_IP_MAP));
            }
            else
            {
                std::pair < std::map<CHANNEL_Plaban, Ipv4Address>::iterator, bool> ret;
                ret = it->second.insert(std::pair<CHANNEL_Plaban, Ipv4Address>(_channel, _ipAdr));
                NS_ASSERT(ret.second);
            }
        }

        void sba_SET_CHNL_CSMA_Mac(int _groupID, CHANNEL_Plaban _channel, Mac48Address _macAdr)
        {
            std::map< int, std::map<CHANNEL_Plaban, Mac48Address> >::iterator it;

            it = this->sba_csmaGroupBasedMacChnlMap.find(_groupID);

            if (this->sba_csmaGroupBasedMacChnlMap.end() == it)
            {
                std::map<CHANNEL_Plaban, Mac48Address> channelMacMAP;
                channelMacMAP.insert(std::pair<CHANNEL_Plaban, Mac48Address>(_channel, _macAdr));

                this->sba_csmaGroupBasedMacChnlMap.insert(std::pair< int, std::map<CHANNEL_Plaban, Mac48Address> >(_groupID, channelMacMAP));
            }
            else
            {
                std::pair < std::map<CHANNEL_Plaban, Mac48Address>::iterator, bool> ret;

                ret = it->second.insert(std::pair<CHANNEL_Plaban, Mac48Address>(_channel, _macAdr));

                if (false == ret.second)
                {// already exists
                    NS_ASSERT(false);
                }
            }
        }

        Mac48Address getCSMA_Mac_givenGroupID_Chnl(int _groupID, CHANNEL_Plaban _channel)
        {
            std::map< int, std::map<CHANNEL_Plaban, Mac48Address> >::iterator it;

            it = this->sba_csmaGroupBasedMacChnlMap.find(_groupID);

            NS_ASSERT(this->sba_csmaGroupBasedMacChnlMap.end() != it); // error if not found

            std::map<CHANNEL_Plaban, Mac48Address>::iterator it2;
            it2 = it->second.find(_channel);

            NS_ASSERT(it->second.end() != it2); // error if not found

            return it2->second;
        }

        void sba_SET_CHNL_WIFI_Mac(int _groupID, CHANNEL_Plaban _channel, Mac48Address _macAdr, Ptr<NetDevice> _netDev)
        {

            ////////////////////////////////////////////////////////////////////////////////////////////////
            std::map< int, std::map<CHANNEL_Plaban, Mac48Address> >::iterator it;

            it = this->sba_wifiAP_GroupBasedMacChnlMap.find(_groupID);

            if (this->sba_wifiAP_GroupBasedMacChnlMap.end() == it)
            {
                std::map<CHANNEL_Plaban, Mac48Address> channelMacMAP;
                channelMacMAP.insert(std::pair<CHANNEL_Plaban, Mac48Address>(_channel, _macAdr));

                this->sba_wifiAP_GroupBasedMacChnlMap.insert(std::pair< int, std::map<CHANNEL_Plaban, Mac48Address> >(_groupID, channelMacMAP));
            }
            else
            {
                std::pair < std::map<CHANNEL_Plaban, Mac48Address>::iterator, bool> ret;

                ret = it->second.insert(std::pair<CHANNEL_Plaban, Mac48Address>(_channel, _macAdr));

                if (false == ret.second)
                {// already exists
                    NS_ASSERT(false);
                }
            }
            /////////////////////////////////////////////////////////////////////////////////////
            //std::map< int, std::map<CHANNEL_Plaban, Ptr<NetDevice> > > sba_wifiAP_GroupBasedChnl_netDev_Map;

            std::map< int, std::map<CHANNEL_Plaban, Ptr<NetDevice> > >::iterator itNetDev;

            itNetDev = this->sba_wifiAP_GroupBasedChnl_netDev_Map.find(_groupID);

            if (this->sba_wifiAP_GroupBasedChnl_netDev_Map.end() == itNetDev)
            {
                std::map<CHANNEL_Plaban, Ptr<NetDevice> > channelNetDevMAP;
                channelNetDevMAP.insert(std::pair<CHANNEL_Plaban, Ptr<NetDevice> >(_channel, _netDev));

                this->sba_wifiAP_GroupBasedChnl_netDev_Map.insert(std::pair< int, std::map<CHANNEL_Plaban, Ptr<NetDevice> > >(_groupID, channelNetDevMAP));
            }
            else
            {
                std::pair < std::map<CHANNEL_Plaban, Ptr<NetDevice> >::iterator, bool> retNetDev;

                retNetDev = itNetDev->second.insert(std::pair<CHANNEL_Plaban, Ptr<NetDevice> >(_channel, _netDev));

                if (false == retNetDev.second)
                {// already exists
                    NS_ASSERT(false);
                }
            }
        }

        Ptr<NetDevice> sba_getWIFI_netDev_givenGroupID_Chnl(int _groupID, CHANNEL_Plaban _channel)
        {
            std::map< int, std::map<CHANNEL_Plaban, Ptr<NetDevice> > >::iterator it;

            it = this->sba_wifiAP_GroupBasedChnl_netDev_Map.find(_groupID);

            NS_ASSERT(this->sba_wifiAP_GroupBasedChnl_netDev_Map.end() != it); // error if not found

            std::map<CHANNEL_Plaban, Ptr<NetDevice> >::iterator it2;
            it2 = it->second.find(_channel);

            NS_ASSERT(it->second.end() != it2); // error if not found

            return it2->second;
        }

        Mac48Address sba_getWIFIMac_givenGroupID_Chnl(int _groupID, CHANNEL_Plaban _channel)
        {
            std::map< int, std::map<CHANNEL_Plaban, Mac48Address> >::iterator it;

            it = this->sba_wifiAP_GroupBasedMacChnlMap.find(_groupID);

            NS_ASSERT(this->sba_wifiAP_GroupBasedMacChnlMap.end() != it); // error if not found

            std::map<CHANNEL_Plaban, Mac48Address>::iterator it2;
            it2 = it->second.find(_channel);

            NS_ASSERT(it->second.end() != it2); // error if not found

            return it2->second;
        }

        void sba_insert_nodeID_MacType_MacADDR_MAP(uint32_t _nodeID, Addr_TYPE_sba _addrType, Mac48Address _macAdr, Ipv4Address _ipAdr)
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////////////
            std::map< uint32_t, std::map<Addr_TYPE_sba, Mac48Address> >::iterator itNode;

            itNode = this->sba_nodeID_macType_MacAdr_MAP.find(_nodeID);

            if (this->sba_nodeID_macType_MacAdr_MAP.end() == itNode)
            {
                std::map<Addr_TYPE_sba, Mac48Address> mactype_macAdr_map;
                mactype_macAdr_map.insert(std::pair<Addr_TYPE_sba, Mac48Address>(_addrType, _macAdr));

                this->sba_nodeID_macType_MacAdr_MAP.insert(std::pair< uint32_t, std::map<Addr_TYPE_sba, Mac48Address> >(_nodeID, mactype_macAdr_map));
            }
            else
            {
                std::pair < std::map<Addr_TYPE_sba, Mac48Address>::iterator, bool > ret;
                ret = itNode->second.insert(std::pair<Addr_TYPE_sba, Mac48Address>(_addrType, _macAdr));

                if (false == ret.second)
                {// already exists
                    NS_ASSERT(false);
                }
            }
            //////////////////////////////////////////////////////////////////////////////////////////////////////////


            NodeID_AdrType nodeIDAddrTypePair(_nodeID, _addrType);

            std::pair < std::map<Ipv4Address, NodeID_AdrType>::iterator, bool> ret;

            ret = this->sba_IP_nodeID_adrType_MAP.insert(std::pair<Ipv4Address, NodeID_AdrType>(_ipAdr, nodeIDAddrTypePair));

            if (false == ret.second)
            {
                NS_ASSERT(false);
            }
        }

        Mac48Address getMACadr_givenNodeID_MacType(uint32_t _nodeID, Addr_TYPE_sba _macType)
        {
            std::map< uint32_t, std::map<Addr_TYPE_sba, Mac48Address> >::iterator itNode;

            itNode = this->sba_nodeID_macType_MacAdr_MAP.find(_nodeID);

            NS_ASSERT(this->sba_nodeID_macType_MacAdr_MAP.end() != itNode); // error if not found

            std::map<Addr_TYPE_sba, Mac48Address>::iterator it2;
            it2 = itNode->second.find(_macType);

            NS_ASSERT(itNode->second.end() != it2); // error if not found

            return it2->second;
        }

        uint32_t sba_getNodeID_givenMAC_Address(Mac48Address _macAdr)
        {
            uint32_t nodeID = 753329;

            std::map< uint32_t, std::map<Addr_TYPE_sba, Mac48Address> >::iterator itNode;

            for (itNode = sba_nodeID_macType_MacAdr_MAP.begin(); sba_nodeID_macType_MacAdr_MAP.end() != itNode; ++itNode)
            {

                std::map<Addr_TYPE_sba, Mac48Address>::iterator itMacTypeAdr;

                for (itMacTypeAdr = itNode->second.begin(); itNode->second.end() != itMacTypeAdr; ++itMacTypeAdr)
                {

                    if (itMacTypeAdr->second == _macAdr)
                    {
                        nodeID = itNode->first;
                        break;
                    }


                }
                if (753329 != nodeID)
                    break;
            }


            return nodeID;

        }

        /**
         * 
         * @param _currentStationNodeID
         * @return the Mac address of the targeted wifi-AP. It will select the targeted wifi AP based on the sender node's current channel.
         */
        //Mac48Address sba_GET_MAC(int _currentStationNodeID, uint32_t _destStationNodeID, Mac48Address& hardwareDestination)

        bool sba_GET_MAC(int _currentStationNodeID, uint32_t _destStationNodeID, Mac48Address& hardwareDestination)
        {// based on its current channel, station will send the packet to one of the Three wifi AP's // assume that we are dealing with three channels
            uint32_t currentStationNodeID = _currentStationNodeID;
            uint32_t destStationNodeID = _destStationNodeID;

            NodeGroupChnlTypeInfo currentNode_groupChnlTypeInfo = this->sba_getNodeInfo_GrpID_currentChnl_nodeType(currentStationNodeID);
            NodeGroupChnlTypeInfo destNode_groupChnlTypeInfo = this->sba_getNodeInfo_GrpID_currentChnl_nodeType(destStationNodeID);

            NS_ASSERT(currentNode_groupChnlTypeInfo.group == destNode_groupChnlTypeInfo.group);

            switch (currentNode_groupChnlTypeInfo.type)
            {
                case AGGREGATOR:
                {
                    NS_ASSERT(NORMAL_NODE == destNode_groupChnlTypeInfo.type);

                    hardwareDestination = this->getCSMA_Mac_givenGroupID_Chnl(currentNode_groupChnlTypeInfo.group, destNode_groupChnlTypeInfo.chnl);
                    return true;

                    break;
                }
                case AP_MODULE:
                {
                    switch (destNode_groupChnlTypeInfo.type)
                    {
                        case NORMAL_NODE:
                        {//flow from AP to STA
                            hardwareDestination = this->getMACadr_givenNodeID_MacType(destStationNodeID, WIFI_ADDR);
                            return true;
                            break;
                        }
                        case AGGREGATOR:
                        {//flow from STA to AP
                            hardwareDestination = this->getMACadr_givenNodeID_MacType(destStationNodeID, CSMA_ADDR);
                            return true;
                            break;
                        }
                        default:
                            return false;
                    }

                    break;
                }
                case NORMAL_NODE:
                {
                    NS_ASSERT(AGGREGATOR == destNode_groupChnlTypeInfo.type);
                    hardwareDestination = this->sba_getWIFIMac_givenGroupID_Chnl(currentNode_groupChnlTypeInfo.group, currentNode_groupChnlTypeInfo.chnl);
                    return true;
                    break;
                }
            }

            return false;

        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

        /**
         * For each Group, there are several WifiAccessPoint modules assigned into a single channel.
         * This function maintains a map where the primary key is the GroupID, and the value is
         * the list of Nodes that are performed as an AP in this particular simulation.
         * @param _gID
         * @param _apID
         */
        void sba_set_group_apNodeList(uint32_t _gID, uint32_t _apID, CHANNEL_Plaban _channel)
        {

            ChannelApID_pair chnlApIDPair(_apID, _channel);

            std::map<int, std::list<ChannelApID_pair> >::iterator it;

            it = this->sba_groupID_ChannelApIDPair_Map.find(_gID);

            if (this->sba_groupID_ChannelApIDPair_Map.end() == it)
            {// not found
                std::list<ChannelApID_pair> apNodeIDList;



                apNodeIDList.push_back(chnlApIDPair);

                this->sba_groupID_ChannelApIDPair_Map.insert(std::pair< uint32_t, std::list<ChannelApID_pair> >(_gID, apNodeIDList));
            }
            else
            {
                it->second.push_back(chnlApIDPair);
            }
        }

        std::list<ChannelApID_pair> sba_get_group_apNodeList(uint32_t _gID)
        {
            std::list<ChannelApID_pair> emptyList;
            emptyList.clear();

            std::map<int, std::list<ChannelApID_pair> >::iterator it;
            it = this->sba_groupID_ChannelApIDPair_Map.find(_gID);

            if (this->sba_groupID_ChannelApIDPair_Map.end() == it)
            {// not found
                return emptyList;
            }
            else
            {
                return it->second;
            }
        }

        void sba_INSERT_NODE_GROUP_CHANNEL(int _nodeID, int _groupID, CHANNEL_Plaban _channel, NodeType_Plaban _type)
        {
            NodeGroupChnlTypeInfo ngctInfo(_groupID, _channel, _type);

            std::pair < std::map<int, NodeGroupChnlTypeInfo>::iterator, bool> ret;
            ret = this->sba_node_group_channel_map.insert(std::pair<int, NodeGroupChnlTypeInfo>(_nodeID, ngctInfo));

            if (false == ret.second)
            {// already exists
                NS_ASSERT(false);
            }

            if (AP_MODULE == _type)
            {
                this->sba_set_group_apNodeList(_groupID, _nodeID, _channel);
            }

        }

        NodeGroupChnlTypeInfo sba_getNodeInfo_GrpID_currentChnl_nodeType(int _nodeID)
        {
            std::map<int, NodeGroupChnlTypeInfo>::iterator it;
            it = this->sba_node_group_channel_map.find(_nodeID);

            if (this->sba_node_group_channel_map.end() == it)
            {
                NS_LOG_UNCOND("\nERROR : node not found !\n");
                NS_ASSERT(false);
            }

            return it->second;
        }

        NodeType_Plaban sba_getNodeType(int _nodeID)
        {
            return this->sba_getNodeInfo_GrpID_currentChnl_nodeType(_nodeID).type;
        }

        CHANNEL_Plaban sba_getNodeCurrentChannel(int _nodeID)
        {
            return this->sba_getNodeInfo_GrpID_currentChnl_nodeType(_nodeID).chnl;
        }

        int sba_getNodeGroup(int _nodeID)
        {
            return this->sba_getNodeInfo_GrpID_currentChnl_nodeType(_nodeID).group;
        }

        bool sba_setNewChannel(int _nodeID, CHANNEL_Plaban _newChnl, Time _currentTime)
        {
            bool isSuccessfull = false;

            std::map<int, NodeGroupChnlTypeInfo>::iterator it;
            it = this->sba_node_group_channel_map.find(_nodeID);

            if (this->sba_node_group_channel_map.end() == it)
            {
                NS_LOG_UNCOND("\nERROR : node not found !\n");
                NS_ASSERT(false);
            }

            Time earliestTime = _currentTime - this->sba_getChannelChangeInterval_inMillisecond();
            Time lastChnlUpdate = it->second.channelLastSetTimeMilliS;
            //            
            //            NS_LOG_UNCOND("\n\t#### 0982424 current Time "<<_currentTime);
            //            NS_LOG_UNCOND("\t#### 0982424 chnlInterval Time "<<this->sba_getChannelChangeInterval_inMillisecond());
            //            NS_LOG_UNCOND("\t#### 0982424 earliest Time "<<earliestTime);
            //            NS_LOG_UNCOND("\t#### 0982424 lastChnlUpdate "<<lastChnlUpdate);
            //            
            //            
            //            NS_LOG_UNCOND("\n\t#### 0982424 current Time ms "<<_currentTime.GetMilliSeconds());
            //            NS_LOG_UNCOND("\t#### 0982424 chnlInterval Time ms "<<this->sba_getChannelChangeInterval_inMillisecond().GetMilliSeconds());
            //            NS_LOG_UNCOND("\t#### 0982424 earliest Time ms "<<earliestTime.GetMilliSeconds());
            //            NS_LOG_UNCOND("\t#### 0982424 lastChnlUpdate ms "<<lastChnlUpdate.GetMilliSeconds());

            Time zeroTime = MilliSeconds(0.0);

            if (-1 == earliestTime.Compare(zeroTime))
            {
                earliestTime = zeroTime;
            }
            //            
            //            NS_LOG_UNCOND("\n\t#### 0982424 current Time ms "<<_currentTime.GetMilliSeconds());
            //            NS_LOG_UNCOND("\t#### 0982424 chnlInterval Time ms "<<this->sba_getChannelChangeInterval_inMillisecond().GetMilliSeconds());
            //            NS_LOG_UNCOND("\t#### 0982424 earliest Time ms "<<earliestTime.GetMilliSeconds());
            //            NS_LOG_UNCOND("\t#### 0982424 lastChnlUpdate ms "<<lastChnlUpdate.GetMilliSeconds());

            if (1 != lastChnlUpdate.Compare(earliestTime))
            {
                //                NS_LOG_UNCOND("\t#### 0982424 True ");
                it->second.chnl = _newChnl;
                it->second.channelLastSetTimeMilliS = _currentTime;
                isSuccessfull = true;
            }

            return isSuccessfull;
        }

        int sba_GET_nodeID_for_IP(Ipv4Address ipv4Adr)
        {
#if 1      

            std::map<Ipv4Address, NodeID_AdrType>::iterator it;

            it = this->sba_IP_nodeID_adrType_MAP.find(ipv4Adr);

            NS_ASSERT(this->sba_IP_nodeID_adrType_MAP.end() != it);

            return it->second.nodeID;

            /*
            int nodeID = -1343;
            
            for(it = this->sba_IP_nodeID_adrType_MAP.begin() ; this->sba_IP_nodeID_adrType_MAP.end() != it ; ++it)
            {
                if(it->first == ipv4Adr)
                {
                    nodeID = it->second.nodeID;
                    break;
                }
            }
            
            return nodeID;
             */

#else 
            std::map<Ipv4Address, int>::iterator it;
            it = this->sba_IP_nodeID_map.find(ipv4Adr);

            if (this->sba_IP_nodeID_map.end() == it)
            {
                NS_LOG_UNCOND("\nERROR : node not found !\n");
                NS_ASSERT(false);
            }

            return it->second;
#endif
        }
        /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

        /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        /**
         * Get the registered TypeId for this class.
         * \return The object TypeId.
         */
        static TypeId GetTypeId(void);

        /**
         * Execute the events scheduled with ScheduleDestroy().
         *
         * This method is typically invoked at the end of a simulation
         * to avoid false-positive reports by a leak checker.
         * After this method has been invoked, it is actually possible
         * to restart a new simulation with a set of calls to Simulator::Run,
         * Simulator::Schedule and Simulator::ScheduleWithContext.
         */
        virtual void Destroy() = 0;
        /**
         * Check if the simulation should finish.
         *
         * Reasons to finish are because there are
         * no more events lefts to be scheduled, or if simulation
         * time has already reached the "stop time" (see Simulator::Stop()).
         *
         * \return \c true if no more events or stop time reached.
         */
        virtual bool IsFinished(void) const = 0;
        /**
         * Tell the Simulator the calling event should be the last one
         * executed.
         *
         * If a running event invokes this method, it will be the last
         * event executed by the Simulator::Run method before
         * returning to the caller.
         */
        virtual void Stop(void) = 0;
        /**
         * Schedule the time delay until the Simulator should stop.
         *
         * Force the Simulator::Run method to return to the caller when the
         * expiration time of the next event to be processed is greater than
         * or equal to the stop time.  The stop time is relative to the
         * current simulation time.
         * \param time The stop time, relative to the current time.
         */
        virtual void Stop(Time const &time) = 0;
        /**
         * Schedule a future event execution (in the same context).
         *
         * \param time Delay until the event expires.
         * \param event The event to schedule.
         * \returns A unique identifier for the newly-scheduled event.
         */
        virtual EventId Schedule(Time const &time, EventImpl *event) = 0;
        /**
         * Schedule a future event execution (in a different context).
         *
         * \param time Delay until the event expires.
         * \param context Event context.
         * \param event The event to schedule.
         * \returns A unique identifier for the newly-scheduled event.
         */
        virtual void ScheduleWithContext(uint32_t context, Time const &time, EventImpl *event) = 0;
        /**
         * Schedule an event to run at the current virtual time.
         *
         * \param event The event to schedule.
         * \returns A unique identifier for the newly-scheduled event.
         */
        virtual EventId ScheduleNow(EventImpl *event) = 0;
        /**
         * Schedule an event to run at the end of the simulation, after
         * the Stop() time or condition has been reached.
         *
         * \param event The event to schedule.
         * \returns A unique identifier for the newly-scheduled event.
         */
        virtual EventId ScheduleDestroy(EventImpl *event) = 0;
        /**
         * Remove an event from the event list. 
         * 
         * This method has the same visible effect as the 
         * ns3::EventId::Cancel method
         * but its algorithmic complexity is much higher: it has often 
         * O(log(n)) complexity, sometimes O(n), sometimes worse.
         * Note that it is not possible to remove events which were scheduled
         * for the "destroy" time. Doing so will result in a program error (crash).
         *
         * \param id The event to remove from the list of scheduled events.
         */
        virtual void Remove(const EventId &id) = 0;
        /**
         * Set the cancel bit on this event: the event's associated function
         * will not be invoked when it expires.
         *
         * This method has the same visible effect as the 
         * ns3::Simulator::Remove method but its algorithmic complexity is 
         * much lower: it has O(1) complexity.
         * This method has the exact same semantics as ns3::EventId::Cancel.
         * Note that it is not possible to cancel events which were scheduled
         * for the "destroy" time. Doing so will result in a program error (crash).
         * 
         * \param id the event to cancel
         */
        virtual void Cancel(const EventId &id) = 0;
        /**
         * Check if an event has already run or been cancelled.
         *
         * This method has O(1) complexity.
         * Note that it is not possible to test for the expiration of
         * events which were scheduled for the "destroy" time. Doing so
         * will result in a program error (crash).
         * An event is said to "expire" when it starts being scheduled
         * which means that if the code executed by the event calls
         * this function, it will get true.
         *
         * \param id The event to test for expiration.
         * \returns \c true if the event has expired, false otherwise.
         */
        virtual bool IsExpired(const EventId &id) const = 0;
        /**
         * Run the simulation.
         *
         * The simulation will run until one of:
         *   - No events are present anymore
         *   - The user called Simulator::Stop
         *   - The user called Simulator::Stop with a stop time and the
         *     expiration time of the next event to be processed
         *     is greater than or equal to the stop time.
         */
        virtual void Run(void) = 0;
        /**
         * Return the current simulation virtual time.
         *
         * \returns The current virtual time.
         */
        virtual Time Now(void) const = 0;
        /**
         * Get the remaining time until this event will execute.
         *
         * \param id The event id to analyse.
         * \return The delay left until the input event id expires.
         *          if the event is not running, this method returns
         *          zero.
         */
        virtual Time GetDelayLeft(const EventId &id) const = 0;
        /**
         * Get the maximum representable simulation time.
         *
         * \return The maximum simulation time at which an event 
         *          can be scheduled.
         *
         * The returned value will always be bigger than or equal to Simulator::Now.
         */
        virtual Time GetMaximumSimulationTime(void) const = 0;
        /**
         * Set the Scheduler to be used to manage the event list.
         *
         * \param schedulerFactory A new event scheduler factory.
         *
         * The event scheduler can be set at any time: the events scheduled
         * in the previous scheduler will be transfered to the new scheduler
         * before we start to use it.
         */
        virtual void SetScheduler(ObjectFactory schedulerFactory) = 0;
        /**
         * Get the system id of this simulator.
         *
         * The system id is the identifier for this simulator instance
         * in a distributed simulation.  For MPI this is the MPI rank.
         * \return The system id for this simulator.
         */
        virtual uint32_t GetSystemId() const = 0;
        /**
         * Get the current simulation context.
         *
         * \return The current simulation context
         */
        virtual uint32_t GetContext(void) const = 0;
    };

} // namespace ns3

#endif /* SIMULATOR_IMPL_H */
