/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2005,2006 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Mathieu Lacage <mathieu.lacage@sophia.inria.fr>
 */
#include "ns3/core-config.h"
#include "simulator.h"
#include "simulator-impl.h"
#include "scheduler.h"
#include "map-scheduler.h"
#include "event-impl.h"

#include "ptr.h"
#include "string.h"
#include "object-factory.h"
#include "global-value.h"
#include "assert.h"
#include "log.h"

#include <cmath>
#include <fstream>
#include <list>
#include <vector>
#include <iostream>

/**
 * \file
 * \ingroup simulator
 * ns3::Simulator implementation, as well as implementation pointer,
 * global scheduler implementation, and default ns3::NodePrinter
 * and ns3::TimePrinter.
 */

namespace ns3
{

    // Note:  Logging in this file is largely avoided due to the
    // number of calls that are made to these functions and the possibility
    // of causing recursions leading to stack overflow
    NS_LOG_COMPONENT_DEFINE("Simulator");

    /**
     * \ingroup simulator
     * The specific simulator implementation to use.
     *
     * Must be derived from SimulatorImpl.
     */
    static GlobalValue g_simTypeImpl = GlobalValue
            ("SimulatorImplementationType",
            "The object class to use as the simulator implementation",
            StringValue("ns3::DefaultSimulatorImpl"),
            MakeStringChecker());

    /**
     * \ingroup scheduler
     * The specific event scheduler implementation to use.
     *
     * Must be derived from Scheduler.
     */
    static GlobalValue g_schedTypeImpl = GlobalValue("SchedulerType",
            "The object class to use as the scheduler implementation",
            TypeIdValue(MapScheduler::GetTypeId()),
            MakeTypeIdChecker());

    /**
     * \ingroup logging
     * Default TimePrinter implementation.
     *
     * \param [in] os The output stream to print the time on.
     */
    static void
    TimePrinter(std::ostream &os)
    {
        os << Simulator::Now().GetSeconds() << "s";
    }

    /**
     * \ingroup logging
     * Default node id printer implementation.
     * 
     * \param [in] os The output stream to print the node id on.
     */
    static void
    NodePrinter(std::ostream &os)
    {
        if (Simulator::GetContext() == 0xffffffff)
        {
            os << "-1";
        }
        else
        {
            os << Simulator::GetContext();
        }
    }

    /**
     * \ingroup simulator
     * \brief Get the static SimulatorImpl instance.
     * \return The SimulatorImpl instance pointer.
     */
    static SimulatorImpl **PeekImpl(void)
    {
        static SimulatorImpl *impl = 0;
        return &impl;
    }

    /**
     * \ingroup simulator
     * \brief Get the SimulatorImpl singleton.
     * \return The singleton pointer.
     * \see Simulator::GetImplementation()
     */
    static SimulatorImpl * GetImpl(void)
    {
        SimulatorImpl **pimpl = PeekImpl();
        /* Please, don't include any calls to logging macros in this function
         * or pay the price, that is, stack explosions.
         */
        if (*pimpl == 0)
        {
            {
                ObjectFactory factory;
                StringValue s;

                g_simTypeImpl.GetValue(s);
                factory.SetTypeId(s.Get());
                *pimpl = GetPointer(factory.Create<SimulatorImpl> ());
            }
            {
                ObjectFactory factory;
                StringValue s;
                g_schedTypeImpl.GetValue(s);
                factory.SetTypeId(s.Get());
                (*pimpl)->SetScheduler(factory);
            }

            //
            // Note: we call LogSetTimePrinter _after_ creating the implementation
            // object because the act of creation can trigger calls to the logging 
            // framework which would call the TimePrinter function which would call 
            // Simulator::Now which would call Simulator::GetImpl, and, thus, get us 
            // in an infinite recursion until the stack explodes.
            //
            LogSetTimePrinter(&TimePrinter);
            LogSetNodePrinter(&NodePrinter);
        }
        return *pimpl;
    }

    void
    Simulator::Destroy(void)
    {
        NS_LOG_FUNCTION_NOARGS();

        SimulatorImpl **pimpl = PeekImpl();
        if (*pimpl == 0)
        {
            return;
        }
        /* Note: we have to call LogSetTimePrinter (0) below because if we do not do
         * this, and restart a simulation after this call to Destroy, (which is 
         * legal), Simulator::GetImpl will trigger again an infinite recursion until
         * the stack explodes.
         */
        LogSetTimePrinter(0);
        LogSetNodePrinter(0);
        (*pimpl)->Destroy();
        (*pimpl)->Unref();
        *pimpl = 0;
    }

    void
    Simulator::SetScheduler(ObjectFactory schedulerFactory)
    {
        NS_LOG_FUNCTION(schedulerFactory);
        GetImpl()->SetScheduler(schedulerFactory);
    }

    bool
    Simulator::IsFinished(void)
    {
        NS_LOG_FUNCTION_NOARGS();
        return GetImpl()->IsFinished();
    }

    void
    Simulator::Run(void)
    {
        NS_LOG_FUNCTION_NOARGS();
        Time::ClearMarkedTimes();
        GetImpl()->Run();
    }

    void
    Simulator::setTestVariable(int a)
    {
        GetImpl()->plaban_test = a;
    }

    int
    Simulator::getTestVariable(void)
    {
        return GetImpl()->plaban_test;
    }

    //    void
    //    Simulator::sba_INSERT_IP_nodeID_map(Ipv4Address ipv4, int nodeID)
    //    {
    //        GetImpl()->sba_INSERT_IP_nodeID_map(ipv4, nodeID);
    //    }

    int
    Simulator::sba_GET_nodeID_for_IP(Ipv4Address ipv4)
    {
        return GetImpl()->sba_GET_nodeID_for_IP(ipv4);
    }


    //    std::map<Ipv4Address, int>
    //    Simulator::sba_getIP_NODE_ID_map()
    //    {
    //        return GetImpl()->sba_getIP_NODE_ID_map();
    //    }

    std::map<Ipv4Address, NodeID_AdrType >
    Simulator::sba_get_IP_nodeID_adrTypeMap()
    {
        return GetImpl()->sba_get_IP_nodeID_adrTypeMap();
    }

    /*
    void
    Simulator::sba_INSERT_MAC_nodeID_map(Mac48Address MAC, int nodeID)
    {
        GetImpl()->sba_INSERT_MAC_nodeID_map(MAC, nodeID);
    }

    int
    Simulator::sba_GET_nodeID_for_MAC(Mac48Address MAC)
    {
        return GetImpl()->sba_GET_nodeID_for_MAC(MAC);
    }
     */
    void
    Simulator::sba_INSERT_NODE_GROUP_CHANNEL(int _nodeID, int _groupID, CHANNEL_Plaban _channel, NodeType_Plaban _type)
    {
        GetImpl()->sba_INSERT_NODE_GROUP_CHANNEL(_nodeID, _groupID, _channel, _type);
    }

    NodeGroupChnlTypeInfo
    Simulator::sba_getNodeInfo_GrpID_currentChnl_nodeType(int _nodeID)
    {
        return GetImpl()->sba_getNodeInfo_GrpID_currentChnl_nodeType(_nodeID);
    }

    NodeType_Plaban
    Simulator::sba_getNodeType(int _nodeID)
    {
        return GetImpl()->sba_getNodeType(_nodeID);
    }

    CHANNEL_Plaban
    Simulator::sba_getNodeCurrentChannel(int _nodeID)
    {
        return GetImpl()->sba_getNodeCurrentChannel(_nodeID);
    }

    int
    Simulator::sba_getNodeGroup(int _nodeID)
    {
        return GetImpl()->sba_getNodeGroup(_nodeID);
    }

    bool
    Simulator::sba_setNewChannel(int _nodeID, CHANNEL_Plaban _newChnl, Time _currentTime)
    {
        return GetImpl()->sba_setNewChannel(_nodeID, _newChnl, _currentTime);
    }

    /*
    void
    Simulator::sba_print_Chnl_Mac_Relation()
    {
        GetImpl()->sba_print_Chnl_Mac_Relation();
    }
     */

    void
    Simulator::sba_SET_CHNL_CSMA_Mac(int _groupID, CHANNEL_Plaban _channel, Mac48Address _macAdr)
    {
        GetImpl()->sba_SET_CHNL_CSMA_Mac(_groupID, _channel, _macAdr);
    }

    void Simulator::sba_SET_CHNL_WIFI_Mac(int _groupID, CHANNEL_Plaban _channel, Mac48Address _macAdr, Ptr<NetDevice> _netDev)
    {
        GetImpl()->sba_SET_CHNL_WIFI_Mac(_groupID, _channel, _macAdr, _netDev);
    }

    void
    Simulator::sba_SET_CHNL_CSMA_IPAdr(int _groupID, CHANNEL_Plaban _channel, Ipv4Address _ipAdr)
    {
        GetImpl()->sba_SET_CHNL_CSMA_IPAdr(_groupID, _channel, _ipAdr);
    }

    void
    Simulator::sba_SET_CHNL_WIFI_IPAdr(int _groupID, CHANNEL_Plaban _channel, Ipv4Address _ipAdr)
    {
        GetImpl()->sba_SET_CHNL_WIFI_IPAdr(_groupID, _channel, _ipAdr);
    }

    //Mac48Address

    bool
    Simulator::sba_GET_MAC(int _currentStationNodeID, uint32_t _destStationNodeID, Mac48Address & hardwareDestination)
    {
        return GetImpl()->sba_GET_MAC(_currentStationNodeID, _destStationNodeID, hardwareDestination);
    }

    Mac48Address
    Simulator::sba_getWIFIMac_givenGroupID_Chnl(int _groupID, CHANNEL_Plaban _channel)
    {
        return GetImpl()->sba_getWIFIMac_givenGroupID_Chnl(_groupID, _channel);
    }

    Ptr<NetDevice>
    Simulator::sba_getWIFI_netDev_givenGroupID_Chnl(int _groupID, CHANNEL_Plaban _channel)
    {
        return GetImpl()->sba_getWIFI_netDev_givenGroupID_Chnl(_groupID, _channel);
    }

    std::map< int, std::map<CHANNEL_Plaban, Mac48Address> >
    Simulator::sba_get_group_chnl_mac_CSMA_MAPPING()
    {
        return GetImpl()->sba_get_group_chnl_mac_CSMA_MAPPING();
    }

    std::map< int, std::map<CHANNEL_Plaban, Mac48Address> >
    Simulator::sba_get_group_chnl_mac_WIFI_MAPPING()
    {
        return GetImpl()->sba_get_group_chnl_mac_WIFI_MAPPING();
    }

    std::map< uint32_t, std::map<Addr_TYPE_sba, Mac48Address> >
    Simulator::get_sba_nodeID_macType_MacAdr_MAP()
    {
        return GetImpl()->get_sba_nodeID_macType_MacAdr_MAP();
    }

    std::map< int, std::map<CHANNEL_Plaban, Ipv4Address> >
    Simulator::sba_get_csmaGroupBased_IPAdr_ChnlMap()
    {
        return GetImpl()->sba_get_csmaGroupBased_IPAdr_ChnlMap();
    }

    std::map< int, std::map<CHANNEL_Plaban, Ipv4Address> >
    Simulator::sba_get_wifiGroupBased_IPAdr_ChnlMap()
    {
        return GetImpl()->sba_get_wifiGroupBased_IPAdr_ChnlMap();
    }

    void
    Simulator::sba_insert_nodeID_AddrType_MacADR_IPAdr_MAP(uint32_t _nodeID, Addr_TYPE_sba _addrType, Mac48Address _macAdr, Ipv4Address _ipAdr)
    {
        GetImpl()->sba_insert_nodeID_MacType_MacADDR_MAP(_nodeID, _addrType, _macAdr, _ipAdr);
    }

    uint32_t Simulator::sba_getNodeID_givenMAC_Address(Mac48Address _macAdr)
    {
        return GetImpl()->sba_getNodeID_givenMAC_Address(_macAdr);
    }

    SIMULATION_TYPE_Plaban
    Simulator::sba_getSimulationType()
    {
        return GetImpl()->sba_getSimulationType();
    }

    void
    Simulator::sba_setSimulationType(SIMULATION_TYPE_Plaban simulationType)
    {
        GetImpl()->sba_setSimulationType(simulationType);
    }

    double
    Simulator::sba_getPktRcvPwrCount(Time _currentTime, int _nodeID, CHANNEL_Plaban channel, int & OUTPUT_receivedPktCount)
    {
        return GetImpl()->sba_getPktRcvPwrCount(_currentTime, _nodeID, channel, OUTPUT_receivedPktCount);
    }

    void
    Simulator::sba_add_rcvd_pkt_info(Time _currentTime, int _nodeID, CHANNEL_Plaban _channel, double rxPwrW, Time rxDuration)
    {
        GetImpl()->sba_add_rcvd_pkt_info(_currentTime, _nodeID, _channel, rxPwrW, rxDuration);
    }

    int
    Simulator::sba_getPktDropCount_onCurrentChnl(Time _currentTime, int _nodeID)
    {
        return GetImpl()->sba_getPktDropCount_onCurrentChnl(_currentTime, _nodeID);
    }

    void
    Simulator::sba_increase_pkt_drop(Time _currentTime, int _nodeID, CHANNEL_Plaban _channel)
    {
        GetImpl()->sba_increase_pkt_drop(_currentTime, _nodeID, _channel);
    }

    void
    Simulator::sba_increase_ReTransmissionCount(Time _currentTime, int _nodeID, CHANNEL_Plaban _channel)
    {
        GetImpl()->sba_increase_ReTransmissionCount(_currentTime, _nodeID, _channel);
    }

    void
    Simulator::sba_setChannelSnsDuration_inMillisecond(int millisecond)
    {
        GetImpl()->sba_setChannelSnsDuration_inMillisecond(millisecond);
    }

    Time
    Simulator::sba_getChannelSnsDuration_inMillisecond()
    {
        return GetImpl()->sba_getChannelSnsDuration_inMillisecond();
    }

    void
    Simulator::sba_setChannelChangeInterval_inMillisecond(int millisecond)
    {
        GetImpl()->sba_setChannelChangeInterval_inMillisecond(millisecond);
    }

    Time
    Simulator::sba_getChannelChangeInterval_inMillisecond()
    {
        return GetImpl()->sba_getChannelChangeInterval_inMillisecond();
    }

    std::map< int, std::map< CHANNEL_Plaban, PktDropCounter> >
    Simulator::sba_get_sba_packetDropCounterPerNodePerChnl_MAP()
    {
        return GetImpl()->sba_get_sba_packetDropCounterPerNodePerChnl_MAP();
    }

    std::map< int, std::map< CHANNEL_Plaban, PktReTxCounter> >
    Simulator::sba_get_sba_packetReTxCounterPerNodePerChnl_MAP()
    {
        return GetImpl()->sba_get_sba_packetReTxCounterPerNodePerChnl_MAP();
    }

    std::map< int, std::map< CHANNEL_Plaban, RcvPktInfoCounter> >
    Simulator::sba_get_sba_pktRcvInfoCounterPerNodePerChnl_MAP()
    {
        return GetImpl()->sba_get_sba_pktRcvInfoCounterPerNodePerChnl_MAP();
    }

    std::list<ChannelApID_pair> Simulator::sba_get_group_apNodeList(uint32_t _gID)
    {
        return GetImpl()->sba_get_group_apNodeList(_gID);
    }

    std::map<NodeType_Plaban, Sta_AP_statistics>
    Simulator::sba_getSta_AP_statisticsMap(Time _currentTime, uint32_t _nodeID)
    {
        return GetImpl()->sba_getSta_AP_statisticsMap(_currentTime, _nodeID);
    }

    void
    Simulator::sba_increaseTotalRetransmissionCount(int _nodeID)
    {
        GetImpl()->sba_increaseTotalRetransmissionCount(_nodeID);
    }

    int
    Simulator::sba_getTotalRetransmissionCount()
    {
        return GetImpl()->sba_getTotalRetransmissionCount();
    }

    void
    Simulator::sba_increaseTotalPktDropCount(int _nodeID)
    {
        GetImpl()->sba_increaseTotalPktDropCount(_nodeID);
    }

    int
    Simulator::sba_getTotalPktDropCount()
    {
        return GetImpl()->sba_getTotalPktDropCount();
    }

    void
    Simulator::Stop(void)
    {
        NS_LOG_FUNCTION_NOARGS();
        NS_LOG_LOGIC("stop");
        GetImpl()->Stop();
    }

    void
    Simulator::Stop(Time const &time)
    {
        NS_LOG_FUNCTION(time);
        GetImpl()->Stop(time);
    }

    Time
    Simulator::Now(void)
    {
        /* Please, don't include any calls to logging macros in this function
         * or pay the price, that is, stack explosions.
         */
        return GetImpl()->Now();
    }

    Time
    Simulator::GetDelayLeft(const EventId &id)
    {
        NS_LOG_FUNCTION(&id);
        return GetImpl()->GetDelayLeft(id);
    }

    EventId
    Simulator::Schedule(Time const &time, const Ptr<EventImpl> &ev)
    {
        return DoSchedule(time, GetPointer(ev));
    }

    EventId
    Simulator::ScheduleNow(const Ptr<EventImpl> &ev)
    {
        return DoScheduleNow(GetPointer(ev));
    }

    void
    Simulator::ScheduleWithContext(uint32_t context, const Time &time, EventImpl *impl)
    {
        return GetImpl()->ScheduleWithContext(context, time, impl);
    }

    EventId
    Simulator::ScheduleDestroy(const Ptr<EventImpl> &ev)
    {
        return DoScheduleDestroy(GetPointer(ev));
    }

    EventId
    Simulator::DoSchedule(Time const &time, EventImpl *impl)
    {
        return GetImpl()->Schedule(time, impl);
    }

    EventId
    Simulator::DoScheduleNow(EventImpl *impl)
    {
        return GetImpl()->ScheduleNow(impl);
    }

    EventId
    Simulator::DoScheduleDestroy(EventImpl *impl)
    {
        return GetImpl()->ScheduleDestroy(impl);
    }

    EventId
    Simulator::Schedule(Time const &time, void (*f)(void))
    {
        return DoSchedule(time, MakeEvent(f));
    }

    void
    Simulator::ScheduleWithContext(uint32_t context, Time const &time, void (*f)(void))
    {
        return ScheduleWithContext(context, time, MakeEvent(f));
    }

    EventId
    Simulator::ScheduleNow(void (*f)(void))
    {
        return DoScheduleNow(MakeEvent(f));
    }

    EventId
    Simulator::ScheduleDestroy(void (*f)(void))
    {
        return DoScheduleDestroy(MakeEvent(f));
    }

    void
    Simulator::Remove(const EventId &id)
    {
        if (*PeekImpl() == 0)
        {
            return;
        }
        return GetImpl()->Remove(id);
    }

    void
    Simulator::Cancel(const EventId &id)
    {
        if (*PeekImpl() == 0)
        {
            return;
        }
        return GetImpl()->Cancel(id);
    }

    bool
    Simulator::IsExpired(const EventId &id)
    {
        if (*PeekImpl() == 0)
        {
            return true;
        }
        return GetImpl()->IsExpired(id);
    }

    Time Now(void)
    {
        return Time(Simulator::Now());
    }

    Time
    Simulator::GetMaximumSimulationTime(void)
    {
        NS_LOG_FUNCTION_NOARGS();
        return GetImpl()->GetMaximumSimulationTime();
    }

    uint32_t
    Simulator::GetContext(void)
    {
        return GetImpl()->GetContext();
    }

    uint32_t
    Simulator::GetSystemId(void)
    {
        NS_LOG_FUNCTION_NOARGS();

        if (*PeekImpl() != 0)
        {
            return GetImpl()->GetSystemId();
        }
        else
        {
            return 0;
        }
    }

    void
    Simulator::SetImplementation(Ptr<SimulatorImpl> impl)
    {
        NS_LOG_FUNCTION(impl);
        if (*PeekImpl() != 0)
        {
            NS_FATAL_ERROR("It is not possible to set the implementation after calling any Simulator:: function. Call Simulator::SetImplementation earlier or after Simulator::Destroy.");
        }
        *PeekImpl() = GetPointer(impl);
        // Set the default scheduler
        ObjectFactory factory;
        StringValue s;
        g_schedTypeImpl.GetValue(s);
        factory.SetTypeId(s.Get());
        impl->SetScheduler(factory);
        //
        // Note: we call LogSetTimePrinter _after_ creating the implementation
        // object because the act of creation can trigger calls to the logging 
        // framework which would call the TimePrinter function which would call 
        // Simulator::Now which would call Simulator::GetImpl, and, thus, get us 
        // in an infinite recursion until the stack explodes.
        //
        LogSetTimePrinter(&TimePrinter);
        LogSetNodePrinter(&NodePrinter);
    }

    Ptr<SimulatorImpl>
    Simulator::GetImplementation(void)
    {
        NS_LOG_FUNCTION_NOARGS();
        return GetImpl();
    }



} // namespace ns3

